# -*- coding: utf-8 -*-
"""
Created on Mon Mar 28 22:27:17 2016

@author: ander
"""
from PyQt5 import QtGui, QtCore, QtWidgets
import sys

def main(args, version=None):

    app = QtWidgets.QApplication(args)
    splash_pix = QtGui.QPixmap('splash_loading.png')
    splash = QtWidgets.QSplashScreen(splash_pix, QtCore.Qt.WindowStaysOnTopHint)
    splash.setMask(splash_pix.mask())
    splash.show()
    app.processEvents()
    app.setOrganizationName("SSHF")
    app.setApplicationName("OpenDXMC {}".format(version))
    import time
    time.sleep(5)
    from opendxmc.app.gui import MainWindow
    win = MainWindow()
    win.show()
    splash.finish(win)

    return app.exec_()



def start(version=None):
    # exit code 1 triggers a restart
    # Also testing for memory error
    try:
        while main(sys.argv, version=version) == 1:
            continue
    except MemoryError:
        msg = QtWidgets.QMessageBox()
        msg.setText("Ouch, OpenDXMC ran out of memory.")
        msg.setIcon(msg.Critical)
        msg.exec_()
    sys.exit(0)


if __name__ == "__main__":
    pass



#
from opendxmc.database.h5database import Database
from opendxmc.database.h5database import PROPETIES_DICT_TEMPLATE, PROPETIES_DICT_TEMPLATE_GROUPING
from opendxmc.database.h5database import Validator
from opendxmc.database.import_phantoms import read_phantoms
from opendxmc.database.import_materials import get_stored_materials
from opendxmc.database.dicom_importer import import_ct_series


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 21 13:39:41 2015

@author: ander
"""

# -*- coding: utf-8 -*-
"""
Created on Mon Dec 14 11:56:58 2015

@author: ander
"""

import ctypes as ct
import numpy as np
import os
import sys
import platform


def get_kernel(precision):
    dll_path = os.path.abspath(os.path.dirname(__file__))
    if platform.system() == 'Windows':
        try:
            if sys.maxsize > 2**32:
                dll = ct.CDLL(os.path.join(dll_path, 'enginelib64'))
            else:
                dll = ct.CDLL(os.path.join(dll_path, 'enginelib32'))
        except OSError:
            if sys.maxsize > 2**32:
                dll = ct.CDLL('enginelib64')
            else:
                dll = ct.CDLL('enginelib32')
    elif platform.system() == 'Linux':
#        dll = ct.CDLL('enginelib.so')
#        print(os.path.join(dll_path, 'enginelib64.so'))
        dll = ct.CDLL(os.path.join(dll_path, 'enginelib64.so'))
    else:
        raise OSError


    setup = dll.setup_simulation
    setup.argtypes = [ct.POINTER(ct.c_int), #shape
                      ct.POINTER(precision), #spacing
                      ct.POINTER(precision), #offset
                      ct.POINTER(ct.c_int32), #materialmap
                      ct.POINTER(precision), #density_map
                      ct.POINTER(ct.c_int32), #lut_shape
                      ct.POINTER(precision), #lut
                      ct.POINTER(precision), #energy_imparted
                      ct.POINTER(ct.c_int32) #use_siddon
                      ]
    setup.restype = ct.c_void_p
    
    source = dll.setup_source
    source.argtypes = [ct.POINTER(precision), #src_pos
                       ct.POINTER(precision), #src_dir
                       ct.POINTER(precision), #scan_axis
                       ct.POINTER(precision), #sdd
                       ct.POINTER(precision), #fov 
                       ct.POINTER(precision), #coll
                       ct.POINTER(precision), #weigh
                       ct.POINTER(precision), #specter cpd
                       ct.POINTER(precision), #specter energy
                       ct.POINTER(ct.c_int32) #specter size
                       ]
    source.restype = ct.c_void_p   
    
    source_bowtie = dll.setup_source_bowtie
    source_bowtie.argtypes = [ct.POINTER(precision), #pos
                              ct.POINTER(precision), #dir
                              ct.POINTER(precision), #scan axis
                              ct.POINTER(precision), #scan angle
                              ct.POINTER(precision), #rot angle
                              ct.POINTER(precision), #weight
                              ct.POINTER(precision), #specter energy
                              ct.POINTER(precision), #spectercpd
                              ct.POINTER(ct.c_int32),    #n_specter
                              ct.POINTER(precision), #bowtie_angle 
                              ct.POINTER(precision), #bowtiw weight
                              ct.POINTER(ct.c_int32)]    #n_bowtie                    
    source_bowtie.restype = ct.c_void_p    
    
    
    run = dll.run_simulation
    run.argtypes = [ct.c_void_p, ct.c_int64, ct.c_void_p]
    run.restype=None    
    
    run_bowtie = dll.run_simulation_bowtie
    run_bowtie.argtypes = [ct.c_void_p, ct.c_int64, ct.c_void_p]    
    run_bowtie.restype=None    
    
    cleanup = dll.cleanup_simulation
    cleanup.argtypes = [ct.c_void_p]
    cleanup.restype=None
   
    cleanup_source = dll.cleanup_source
    cleanup_source.argtypes = [ct.c_void_p]
    cleanup_source.restype=None
    #info = dll.device_info
    
    return setup, source, source_bowtie, run, run_bowtie, cleanup, cleanup_source

class Engine(object):
    def __init__(self, precision='float64'):
        if precision == 'float64':
            self.floating_type = ct.c_double
        else:
            self.floating_type = ct.c_float
        self.c_simsetup, self.c_sourcesetup, self.c_sourcesetup_bowtie, self.crun, self.crun_bowtie, self.c_simcleanup, self.c_sourcecleanup = get_kernel(self.floating_type)

    def setup_simulation(self, shape, spacing, offset, material_map, density_map, lut_shape, lut, energy_imparted, use_siddon):
        return self.c_simsetup(
                 shape.ctypes.data_as(ct.POINTER(ct.c_int32)), 
                 spacing.ctypes.data_as(ct.POINTER(self.floating_type)),
                 offset.ctypes.data_as(ct.POINTER(self.floating_type)), 
                 material_map.ctypes.data_as(ct.POINTER(ct.c_int32)), 
                 density_map.ctypes.data_as(ct.POINTER(self.floating_type)), 
                 lut_shape.ctypes.data_as(ct.POINTER(ct.c_int32)), 
                 lut.ctypes.data_as(ct.POINTER(self.floating_type)), 
                 energy_imparted.ctypes.data_as(ct.POINTER(self.floating_type)),
                 use_siddon.ctypes.data_as(ct.POINTER(ct.c_int32)), 
                 ) 

                 
    def setup_source(self, source_position, source_direction, scan_axis, sdd, 
                     fov, collimation, weight, specter_cpd, specter_energy, n_specter):
        
        return self.c_sourcesetup(
                    source_position.ctypes.data_as(ct.POINTER(self.floating_type)),
                    source_direction.ctypes.data_as(ct.POINTER(self.floating_type)),
                    scan_axis.ctypes.data_as(ct.POINTER(self.floating_type)),
                    sdd.ctypes.data_as(ct.POINTER(self.floating_type)),
                    fov.ctypes.data_as(ct.POINTER(self.floating_type)),
                    collimation.ctypes.data_as(ct.POINTER(self.floating_type)),
                    weight.ctypes.data_as(ct.POINTER(self.floating_type)),
                    specter_cpd.ctypes.data_as(ct.POINTER(self.floating_type)),
                    specter_energy.ctypes.data_as(ct.POINTER(self.floating_type)),
                    n_specter.ctypes.data_as(ct.POINTER(ct.c_int32))
                    )
    def setup_source_bowtie(self, source_position, source_direction, scan_axis, 
                            scan_angle, rot_angle, weight, specter_cpd,
                            specter_energy, n_specter, bowtie_weight, bowtie_angle, n_bowtie):
                                
        
#        import pdb ; pdb.set_trace()
        return self.c_sourcesetup_bowtie(
                    source_position.ctypes.data_as(ct.POINTER(self.floating_type)),
                    source_direction.ctypes.data_as(ct.POINTER(self.floating_type)),
                    scan_axis.ctypes.data_as(ct.POINTER(self.floating_type)),
                    scan_angle.ctypes.data_as(ct.POINTER(self.floating_type)),
                    rot_angle.ctypes.data_as(ct.POINTER(self.floating_type)),
                    weight.ctypes.data_as(ct.POINTER(self.floating_type)),
                    specter_cpd.ctypes.data_as(ct.POINTER(self.floating_type)),
                    specter_energy.ctypes.data_as(ct.POINTER(self.floating_type)),
                    n_specter.ctypes.data_as(ct.POINTER(ct.c_int32)),
                    bowtie_weight.ctypes.data_as(ct.POINTER(self.floating_type)),
                    bowtie_angle.ctypes.data_as(ct.POINTER(self.floating_type)),
                    n_bowtie.ctypes.data_as(ct.POINTER(ct.c_int32))
                    )
    def run(self, source_ptr, n_particles, sim_ptr):
        self.crun(source_ptr, 
                   ct.c_int64(n_particles), 
                   sim_ptr)
    def run_bowtie(self, source_ptr, n_particles, sim_ptr):
        self.crun_bowtie(source_ptr, 
                         ct.c_int64(n_particles), 
                         sim_ptr)
        
    def cleanup(self, simulation=None,source=None):
        if simulation:
            self.c_simcleanup(simulation)
        if source:
            self.c_sourcecleanup(source)
            
            
#            
#def test():
#    e = get_kernel('float64')
#    print(e)
#    
#    
#def test():
#    import matplotlib.pylab as plt
#    res = 64
#    material_map = np.zeros((res, res, res), dtype="int")
#    density_map = np.ones((res, res, res), dtype="float64")
#    
#    energy_imparted = np.zeros((res, res, res), dtype="float64")
#    
#    # lut    
#    lut = np.zeros((2, 5, 5), dtype="float64")
#    lut[0, 0, :] = np.array([1000, 10000, 50000, 69000, 100000])
#    lut[0, 2, :] = np.array([.34, .0246, 0.00101, .00005, .000276])
#    lut[0, 3, :] = np.array([6.8, .00277, .000011, .000003, .000000987])
#    lut[0, 4, :] = np.array([.05, .385, .3344, .317, .29])
#    for i in range(lut.shape[2]):
#        lut[0, 1, i] = lut[0, 2:, i].sum()
#     
#    shape = np.array(material_map.shape, dtype="int")
#    spacing = np.array([.1, .1, .1], dtype="float64")
#    offset = (-shape*spacing / 2. ).astype("float64")
#    lut_shape = np.array(lut.shape, dtype="int")
#     
#         
#     
#    s_pos = np.array([100, 0, 0], dtype='float64')
#    s_dir = np.array([-1, 0, 0], dtype='float64')
#    axis = np.array([0, 0, 1], dtype='float64')
#    fov = np.array([50,], dtype='float64')
#    sdd = np.array([100,], dtype='float64')
#    collimation = np.array([4,], dtype='float64')
#    weight = np.array([1,], dtype='float64')
#    specter_cpd = np.cumsum(np.array([1, 1, 1], dtype='float64'))
#    specter_cpd /= specter_cpd.max()
#    specter_energy = np.array([60, 70, 80], dtype='float64')
#    n_specter = np.array([3], dtype='int')
#    
#    # test_setup
#    sim_setup, source_setup, kernel_run, sim_cleanup, source_cleanup = get_kernel()
#    
#    
#   
##    pdb.set_trace()
#
#    timer = time.clock()
#
##    pdb.set_trace()
#    sim_ptr = sim_setup(
#                 shape.ctypes.data_as(ct.POINTER(ct.c_int)), 
#                 spacing.ctypes.data_as(ct.POINTER(ct.c_double)),
#                 offset.ctypes.data_as(ct.POINTER(ct.c_double)), 
#                 material_map.ctypes.data_as(ct.POINTER(ct.c_int)), 
#                 density_map.ctypes.data_as(ct.POINTER(ct.c_double)), 
#                 lut_shape.ctypes.data_as(ct.POINTER(ct.c_int)), 
#                 lut.ctypes.data_as(ct.POINTER(ct.c_double)), 
#                 energy_imparted.ctypes.data_as(ct.POINTER(ct.c_double))
#                 )
#
#    source_ptr = source_setup(
#                    s_pos.ctypes.data_as(ct.POINTER(ct.c_double)),
#                    s_dir.ctypes.data_as(ct.POINTER(ct.c_double)),
#                    axis.ctypes.data_as(ct.POINTER(ct.c_double)),
#                    sdd.ctypes.data_as(ct.POINTER(ct.c_double)),
#                    fov.ctypes.data_as(ct.POINTER(ct.c_double)),
#                    collimation.ctypes.data_as(ct.POINTER(ct.c_double)),
#                    weight.ctypes.data_as(ct.POINTER(ct.c_double)),
#                    specter_cpd.ctypes.data_as(ct.POINTER(ct.c_double)),
#                    specter_energy.ctypes.data_as(ct.POINTER(ct.c_double)),
#                    n_specter.ctypes.data_as(ct.POINTER(ct.c_int))
#                    )
#
#    
#    #void_ptr_c = ct.c_void_p(void_ptr)
#    print('Setup time', time.clock()-timer)
#    
#    n_particles = 5000000
#            
#    teller = 0
#    timer_tot = time.clock()
#    while True:
#        
#        timer = time.clock()
#        kernel_run(source_ptr, ct.c_size_t(n_particles), sim_ptr)
#        print('histories per second', n_particles / (time.clock()-timer), time.clock()-timer)
#        teller += 1
#        if teller > 5:
#            break
#    print("total time for {} histories is {}".format((teller)*n_particles, time.clock()-timer_tot))
#        
#                   
#
#    timer = time.clock()
#    sim_cleanup(sim_ptr, 
#                shape.ctypes.data_as(ct.POINTER(ct.c_int)), 
#                energy_imparted.ctypes.data_as(ct.POINTER(ct.c_double))) 
#    source_cleanup(source_ptr)
#    print('cleanup time', time.clock()-timer)
#    print(shape)
#    print('max value', energy_imparted[:, : , res//2].max())
#    cut = energy_imparted[:, : , res//2].max()/10000.
#    cut = 100000
#    print('cut', cut)
#    plt.subplot(121)
#    plt.imshow(np.clip(energy_imparted[:, : , res//2], 0,cut))
#    plt.subplot(122)
#    plt.imshow(energy_imparted[:, : , res//2])
#    plt.show()           
#if __name__ == '__main__':
#    test()

    
    




# -*- coding: utf-8 -*-
"""
Created on Tue Oct 13 11:03:25 2015

@author: erlean
"""

import os
import sys
import zipfile
import numpy as np
from opendxmc.database.h5database import Validator
from opendxmc.data.phantom_definitions import golem_organs, vishum_organs, donna_organs, helga_organs, irene_organs, eva_organs, adam_organs, frank_organs, katja_organs, child_organs, baby_organs, jo_organs
import logging
logger = logging.getLogger('OpenDXMC')


PATH = os.path.join(os.path.dirname(sys.argv[0]), 'opendxmc', 'data', 'phantoms')

def read_phantoms(file_list):

    phantoms = {'golem': ('segm_golem', golem_organs, (2.08, 2.08, 8.), (256, 256, 220)),
                'VisHum': ('segm_vishum', vishum_organs, (.91, .94, 5.), (512, 512, 250)),
                'Donna':  ('segm_donna', donna_organs, (1.875, 1.875, 10.), (256, 256, 179)),
                'Helga':  ('segm_helga', helga_organs, (.98, .98, 10.), (512, 512, 114)),
                'Irene':  ('Irene', irene_organs, (1.875, 1.875, 5.), (132, 262, 348)),
                'Eva8bit': ('EVA_8bit', eva_organs, (1.6, 1.6, 2.), (256, 256, 900)),
                'Adam8bit': ('ADAM_8bit', adam_organs, (1.6, 1.6, 2.), (256, 256, 900)),
                'Frank': ('segm_frank', frank_organs, (.742, .742, 5.), (512, 512, 193)),
                'Katja': ('Katja', katja_organs, (1.775, 1.775, 4.84), (150, 299, 348)),
                'Child': ('segm_child', child_organs, (1.54, 1.54, 8), (256, 256, 144)),
                'Babynew': ('babynew_May2003', baby_organs, (.85, .85, 4.), (138, 267, 142)),
                'Jo': ('Jo', jo_organs, (1.875, 1.875, 10.), (118, 226, 136)),
                }


    unzipped = False
    for fil in file_list:
        path = os.path.abspath(fil)
        if zipfile.is_zipfile(fil):
            logger.debug('Attempting to unzip file {}'.format(path))
            with zipfile.ZipFile(path) as fz:
                for phantom, values in phantoms.items():
                    try:
                        path = os.path.abspath(fz.extract(values[0], path=os.path.dirname(path)))
                    except:
                        pass
                    else:
                        logger.debug('File {0} unzipped to {1}'.format(values[0], path))
                        unzipped = True
                        break
        filename = os.path.basename(path)
        for key, value in phantoms.items():
            if filename == value[0]:
                try:
                    sim = read_voxels(path, key, *value[1:])
                except FileNotFoundError:
                    logger.debug('{0} phantom not found in {1}'.format(key, path))
                else:
                    logger.debug('{0} phantom read successfully'.format(key))
                    yield sim.get_data()
                if unzipped:
                    try:
                        os.remove(path)
                    except:
                        pass
                unzipped = False


#    for key, value in phantoms.items():
#        path = os.path.join(PATH, key, value[0])
#        try:
#            sim = read_voxels(path, key, *value[1:])
#        except FileNotFoundError as e:
#            logger.debug('{0} phantom not found in {1}'.format(key, path))
##            raise e
#        else:
#            logger.debug('{0} phantom read successfully'.format(key))
#
#            yield sim.get_data()



def read_voxels(path, name, organ_func, spacing, shape, header_len=4096):
    voxpath = path
    logger.debug('Attempting to read {0} phantom from {1}'.format(name, voxpath))

    a = np.fromfile(voxpath, dtype=np.uint8)
    sim = Validator()
    sim.name = name
    sim.spacing = np.array(spacing, dtype=np.double) / 10.
    sim.organ = np.rollaxis(np.reshape(a[header_len:], (-1, shape[0], shape[1])), 0, 3)#[:,:,::-1]
    sim.shape = np.array(sim.organ.shape)
    N = sim.organ.shape[2]
    organ_map={}
    organ_material_map={}
    organ_values = []
    for i in range(N):
        # we do loops due to memory constrains on np.unique
        organ_values += list(np.unique(sim.organ[:, : ,i]))

    organ_dict = {organ_number: (organ, tissue) for organ_number, organ, tissue in organ_func()}
    for organ_number in set(organ_values):
        if organ_number in organ_dict:
            organ_str, tissue_str = organ_dict[organ_number]
            organ_map[organ_number] = organ_str
            organ_material_map[organ_number] = tissue_str
        else:
            organ_map[organ_number] = 'Unknown as air {}'.format(organ_number)
            organ_material_map[organ_number] = 'air'

    sim.organ_map = organ_map
    sim.organ_material_map = organ_material_map
    sim.is_phantom = True
    sim.start_scan = 0
    sim.stop_scan = sim.spacing[2] * N
    sim.stop = sim.spacing[2] * N
    sim.data_center = np.array(sim.organ.shape) * sim.spacing /2.
    aec = np.ones((N, 2))
    aec[:, 0] = np.linspace(0, sim.stop, N)

    sim.exposure_modulation = aec


    sim.scaling=np.array([1,1,1], dtype=np.double)
    return sim





# -*- coding: utf-8 -*-
"""
Created on Thu Aug 27 14:41:06 2015

@author: erlean
"""


import os
import sys
import logging
import re
from opendxmc.materials.materials import Material
from opendxmc.utils import find_all_files

logger = logging.getLogger('OpenDXMC')

MATERIAL_DATA_PATH = os.path.join(os.path.dirname(sys.argv[0]), 'opendxmc',
                                  'data', 'materials')

BASE_PATH = os.path.dirname(os.path.dirname(__file__))

MATERIAL_DATA_PATH = os.path.join(BASE_PATH,
                                  'data', 'materials')


def get_stored_materials(material_data_path=None):
    if material_data_path is None:
        material_data_path = MATERIAL_DATA_PATH
    logger.debug('Importing materials from {}'.format(material_data_path))
    density_file = os.path.join(material_data_path, "densities.txt")
    organic_file = os.path.join(material_data_path, "organics.txt")

    for p in find_all_files([os.path.join(material_data_path, 'attenuation')]):
        name = str(os.path.splitext(os.path.basename(p))[0])
        # test for valid material name
        if re.match('^[\w-]+$', name) is None:
            logger.warning(
                "Material file {0} contains illegal characters"
                ", only alphanummeric characters and "
                "dashes are allowed.".format(p)
                )
            continue
        try:
            yield Material(name, att_file=p, density_file=density_file,
                           organic_file=organic_file)
        except Exception as e:
            logger.warning(str(e))
    return



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 15:15:30 2015

@author: erlean
"""
import numpy as np
import logging

import pdb

logger = logging.getLogger('OpenDXMC')


class Material(object):
    """
    Material object for interaction with database.

    INPUT:

    name:
        Name of material, only alphanummeric values are allowed
    density [optional]:
        material density in g/cm3
    att_file [optional]:
        filepath to text file containing attinuation data
    attinuations [optional]:
        attinuation nd array
    density_file [optional]:
        read density from this file

    """
    def __init__(self, name, density=None, organic=None, att_file=None,
                 attinuations=None, density_file=None, organic_file=None):
        self.name = name
        self.__density = density
        self.__atts = attinuations
        self.__organic = organic

        if att_file is not None:
            self.attinuation = att_file
        if density_file is not None:
            self.density_from_file(density_file)
        if organic_file is not None:
            self.organic_from_file(organic_file)

    def numpy_dtype(self):
        return np.dtype({'names': ['name', 'density', 'organic'],
                         'formats': ['a64', np.double, np.bool]})
    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        if isinstance(value, bytes):
            value = str(value, encoding='utf-8')
        else:
            value = str(value)
        name = "".join([l for l in value.split() if len(l) > 0])
        assert len(name) > 0
        self.__name = name.lower()

    @property
    def density(self):
        return self.__density

    @density.setter
    def density(self, value):
        self.__density = float(value)

    @property
    def organic(self):
        return self.__organic

    @organic.setter
    def organic(self, value):
        self.__organic = bool(value)

    def organic_from_file(self, path):
        try:
            with open(path) as f:
                s = f.read()
                l = s.lower().split()
                if self.name in l:
                    self.organic = True
        except FileNotFoundError:
            logger.warning("Could not open organic file {0} "
                           "for material {1}".format(path, self.name))

    def density_from_file(self, path):
        try:
            with open(path) as f:
                s = f.read()
                l = s.lower().split()
                if self.name in l:
                    self.density = l[l.index(self.name) + 1]
        except FileNotFoundError:
            logger.warning("Could not open densities file {0} "
                           "for material {1}".format(path, self.name))

    @property
    def attinuation(self):
        return self.__atts

    @attinuation.setter
    def attinuation(self, path):
        letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        letters += letters.lower()
        header = ('energy', 'rayleigh', 'compton', 'photoelectric', 'ppn',
                  'ppe', 'total', 'totalnocoherent')
        data_list = dict([(h, list()) for h in header])
        with open(path) as f:
            # reading lines
            for line in f:
                line = line.strip()
                # testing for header or empty line
                if len(line) == 0:
                    continue

                if line[0] in letters:
                    continue
                for i, el in enumerate(line.split()):
                    data_list[header[i]].append(float(el))
        N = len(data_list['energy'])
        dtype = [(h, np.double) for h in header]
        array = np.recarray(N, dtype=dtype)
        for h in header:
            array[h] = np.array(data_list[h], dtype=np.double)[:]
        array['energy'] *= 1.e6
        array.sort(order='energy')
        self.__atts = array



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 30 10:38:15 2015

@author: erlean
"""

import numpy as np
from PyQt5 import QtGui, QtCore, QtWidgets
import os
import pyqtgraph
import pyqtgraph.opengl as gl
from OpenGL.GL import glReadPixels, GL_RGBA, GL_UNSIGNED_BYTE, GL_RGB
from .dicom_lut import get_lut, get_lut_raw
from scipy.ndimage.filters import gaussian_filter
from opendxmc.app.ffmpeg_writer import FFMPEG_VideoWriter
import logging
logger = logging.getLogger('OpenDXMC')


## @file
#
# Qt Widget Range slider widget. This is a slider with two bars, one of
# which is used to set the maximum and the other of which is used to
# set the minimum.
#
# Qt Double Spin Box slider widget. This is the range slider grouped
# with two spin boxes to make the range editable.
#
# Hazen 07/14
#

class QRangeSlider(QtWidgets.QWidget):
    doubleClick = QtCore.pyqtSignal(bool)
    rangeChanged = QtCore.pyqtSignal(float, float)

    ## __init__
    #
    # @param slider_range [min, max, step size].
    # @param values [initial minimum setting, initial maximum setting].
    # @param parent (Optional) The PyQt parent of this widget.
    #
    def __init__(self, slider_range, values, parent = None):
        super().__init__(parent)
        self.bar_width = 10
        self.emit_while_moving = False
        self.moving = "none"
        self.old_scale_min = 0.0
        self.old_scale_max = 0.0
        self.scale = 0
        self.setMouseTracking(False)
        self.single_step = 0.0

        if slider_range:
            self.setRange(slider_range)
        else:
            self.setRange([0.0, 1.0, 0.01])
        if values:
            self.setValues(values)
        else:
            self.setValues([0.3, 0.6])

        self.setFocusPolicy(QtCore.Qt.ClickFocus)

    def emitRange(self):
        if (self.old_scale_min != self.scale_min) or (self.old_scale_max != self.scale_max):
            self.rangeChanged.emit(self.scale_min, self.scale_max)
            self.old_scale_min = self.scale_min
            self.old_scale_max = self.scale_max

    def getValues(self):
        return [self.scale_min, self.scale_max]

    def keyPressEvent(self, event):
        key = event.key()

        # move bars based on arrow keys
        moving_max = False
        if key == QtCore.Qt.Key_Up:
            self.scale_max += self.single_step
            moving_max = True
        elif key == QtCore.Qt.Key_Down:
            self.scale_max -= self.single_step
            moving_max = True
        elif key == QtCore.Qt.Key_Left:
            self.scale_min -= self.single_step
        elif key == QtCore.Qt.Key_Right:
            self.scale_min += self.single_step

        # update (if necessary) based on allowed range
        if moving_max:
            if (self.scale_max < self.scale_min):
                self.scale_min = self.scale_max
        else:
            if (self.scale_min > self.scale_max):
                self.scale_max = self.scale_min

        if (self.scale_min < self.start):
            self.scale_min = self.start
        if (self.scale_max < self.start):
            self.scale_max = self.start

        slider_max = self.start + self.scale
        if (self.scale_min > slider_max):
            self.scale_min = slider_max
        if (self.scale_max > slider_max):
            self.scale_max = slider_max

        self.emitRange()
        self.updateDisplayValues()
        self.update()


    def mouseDoubleClickEvent(self, event):
        self.doubleClick.emit(True)

    def mouseMoveEvent(self, event):
        size = self.rangeSliderSize()
        diff = self.start_pos - self.getPos(event)
        if self.moving == "min":
            temp = self.start_display_min - diff
            if (temp >= self.bar_width) and (temp < size - self.bar_width):
                self.display_min = temp
                if self.display_max < self.display_min:
                    self.display_max = self.display_min
                self.updateScaleValues()
                if self.emit_while_moving:
                    self.emitRange()
        elif self.moving == "max":
            temp = self.start_display_max - diff
            if (temp >= self.bar_width) and (temp < size - self.bar_width):
                self.display_max = temp
                if self.display_max < self.display_min:
                    self.display_min = self.display_max
                self.updateScaleValues()
                if self.emit_while_moving:
                    self.emitRange()
        elif self.moving == "bar":
            temp = self.start_display_min - diff
            if (temp >= self.bar_width) and (temp < size - self.bar_width - (self.start_display_max - self.start_display_min)):
                self.display_min = temp
                self.display_max = self.start_display_max - diff
                self.updateScaleValues()
                if self.emit_while_moving:
                    self.emitRange()

    def mousePressEvent(self, event):
        pos = self.getPos(event)
        if abs(self.display_min - 0.5 * self.bar_width - pos) < (0.5 * self.bar_width):
            self.moving = "min"
        elif abs(self.display_max + 0.5 * self.bar_width - pos) < (0.5 * self.bar_width):
            self.moving = "max"
        elif (pos > self.display_min) and (pos < self.display_max):
            self.moving = "bar"
        self.start_display_min = self.display_min
        self.start_display_max = self.display_max
        self.start_pos = pos

    def mouseReleaseEvent(self, event):
        if not (self.moving == "none"):
            self.emitRange()
        self.moving = "none"

    def resizeEvent(self, event):
        self.updateDisplayValues()

    def setEmitWhileMoving(self, flag):
        if flag:
            self.emit_while_moving = True
        else:
            self.emit_while_moving = False

    def setRange(self, slider_range):
        self.start = slider_range[0]
        self.scale = slider_range[1] - slider_range[0]
        self.single_step = slider_range[2]

        # Check that the range is a multiple of the step size.
        steps = self.scale / self.single_step
        if (abs(steps - round(steps)) > 0.01 * self.single_step):
            self.single_step = steps


    def setValues(self, values):
        self.scale_min = values[0]
        self.scale_max = values[1]
        self.emitRange()
        self.updateDisplayValues()
        self.update()

    def updateDisplayValues(self):
        size = float(self.rangeSliderSize() - 2 * self.bar_width - 1)
        self.display_min = int(size * (self.scale_min - self.start)/self.scale) + self.bar_width
        self.display_max = int(size * (self.scale_max - self.start)/self.scale) + self.bar_width

    def updateScaleValues(self):
        size = float(self.rangeSliderSize() - 2 * self.bar_width - 1)
        if (self.moving == "min") or (self.moving == "bar"):
            self.scale_min = self.start + (self.display_min - self.bar_width)/float(size) * self.scale
            self.scale_min = float(round(self.scale_min/self.single_step))*self.single_step
        if (self.moving == "max") or (self.moving == "bar"):
            self.scale_max = self.start + (self.display_max - self.bar_width)/float(size) * self.scale
            self.scale_max = float(round(self.scale_max/self.single_step))*self.single_step
        self.updateDisplayValues()
        self.update()



class QHRangeSlider(QRangeSlider):

    ## __init__
    #
    # @param slider_range (Optional) [min, max, step size].
    # @param values (Optional) [initial minimum setting, initial maximum setting].
    # @param parent (Optional) The PyQt parent of this widget.
    #
    def __init__(self, slider_range = None, values = None, parent = None):
        super().__init__(slider_range, values, parent)
        if (not parent):
            self.setGeometry(200, 200, 200, 100)

    def getPos(self, event):
        return event.x()

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        w = self.width()
        h = self.height()

        # background
        painter.setPen(QtCore.Qt.gray)
        painter.setBrush(QtCore.Qt.lightGray)
        painter.drawRect(2, 2, w-4, h-4)

        # range bar
        painter.setPen(QtCore.Qt.darkGray)
        painter.setBrush(QtCore.Qt.darkGray)
        painter.drawRect(self.display_min-1, 5, self.display_max-self.display_min+2, h-10)

        # min & max tabs
        painter.setPen(QtCore.Qt.black)
        painter.setBrush(QtCore.Qt.gray)
        painter.drawRect(self.display_min-self.bar_width, 1, self.bar_width, h-2)

        painter.setPen(QtCore.Qt.black)
        painter.setBrush(QtCore.Qt.gray)
        painter.drawRect(self.display_max, 1, self.bar_width, h-2)

    def rangeSliderSize(self):
        return self.width()

class QVRangeSlider(QRangeSlider):

    ## __init__
    #
    # @param slider_range (Optional) [min, max, step size].
    # @param values (Optional) [initial minimum setting, initial maximum setting].
    # @param parent (Optional) The PyQt parent of this widget.
    #
    def __init__(self, slider_range = None, values = None, parent = None):
        super().__init__(slider_range, values, parent)
        if (not parent):
            self.setGeometry(200, 200, 100, 200)
    def getPos(self, event):
        return self.height() - event.y()

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        w = self.width()
        h = self.height()

        # background
        painter.setPen(QtCore.Qt.gray)
        painter.setBrush(QtCore.Qt.lightGray)
        painter.drawRect(2, 2, w-4, h-4)

        # range bar
        painter.setPen(QtCore.Qt.darkGray)
        painter.setBrush(QtCore.Qt.darkGray)
        painter.drawRect(5, h-self.display_max-1, w-10, self.display_max-self.display_min+1)

        # min & max tabs
        painter.setPen(QtCore.Qt.black)
        painter.setBrush(QtCore.Qt.gray)
        painter.drawRect(1, h-self.display_max-self.bar_width-1, w-2, self.bar_width)

        painter.setPen(QtCore.Qt.black)
        painter.setBrush(QtCore.Qt.gray)
        painter.drawRect(1, h-self.display_min-1, w-2, self.bar_width)

    ## rangeSliderSize
    #
    # @return The current height of the slider widget.
    #
    def rangeSliderSize(self):
        return self.height()


class SceneSelectGroup(QtWidgets.QActionGroup):
    scene_selected = QtCore.pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setExclusive(True)
        self.triggered.connect(self.relay_clicked)

    def addAction(self, name, pretty_name=None):
        if pretty_name is None:
            pretty_name = name
        action = super().addAction(pretty_name)
        action.scene_name = name

    @QtCore.pyqtSlot(QtGui.QAction)
    def relay_clicked(self, action):
        self.scene_selected.emit(action.scene_name)

    @QtCore.pyqtSlot(str)
    def sceneSelected(self, name):
        for action in self.actions():
            if action.scene_name == name:
                action.setChecked()
                return

class VolumeViewManager(QtWidgets.QWidget):
    def __init__(self, volume_scenes, parent=None):
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        main_layout = QtWidgets.QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setAlignment(QtCore.Qt.AlignTop)

        menu_layout = QtWidgets.QHBoxLayout()
        menu_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.addLayout(menu_layout)

        enable_vrt_button = QtWidgets.QPushButton('Enable VRTs')
        enable_vrt_button.setCheckable(True)
        enable_vrt_button.setFlat(True)
        enable_vrt_button.setToolTip('Enable VRTs if your graphics card is up for it')
        enable_vrt_button.setStatusTip('Enable VRTs if your graphics card is up for it')
        menu_layout.addWidget(enable_vrt_button)

        splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        main_layout.addWidget(splitter)

        button_icon = QtGui.QIcon(os.path.join(os.path.dirname(__file__), 'copy_icon.png'))

        for name, wid3d in volume_scenes.items():
            enable_vrt_button.toggled.connect(wid3d.set_showing)
            wid = QtWidgets.QWidget()
            wid.setVisible(False)
            wid3d.is_showing.connect(wid.setVisible)

            v_lay = QtWidgets.QHBoxLayout()
            v_lay.setAlignment(QtCore.Qt.AlignRight)
            v_lay.setContentsMargins(0, 0, 0, 0)
            v_lay.addWidget(wid3d)

            ranges = list(wid3d.custom_data_range)
            ranges += [(ranges[1]-ranges[0])/25,]

            slider = QVRangeSlider(ranges, [ranges[0], ranges[1]])

            slider.setMinimumWidth(20)
            slider.setMaximumWidth(20)
            slider.rangeChanged.connect(wid3d.set_custom_data_range)
            l_lay = QtWidgets.QVBoxLayout()
            l_lay.addWidget(slider)

            copy_button = QtWidgets.QPushButton()
            copy_button.setFlat(True)
            copy_button.setToolTip('Copy VRT image to clipboard')
            copy_button.setIcon(button_icon)
            copy_button.clicked.connect(wid3d.save_image_to_clipboard)
            l_lay.addWidget(copy_button)

            v_lay.addLayout(l_lay)
            wid.setLayout(v_lay)
            wid.setContentsMargins(0, 0, 0, 0)
            splitter.addWidget(wid)
        enable_vrt_button.clicked.connect(self.update)
        self.setLayout(main_layout)


class ViewController(QtCore.QObject):
    request_metadata = QtCore.pyqtSignal(str)
    request_array_slice = QtCore.pyqtSignal(str, str, int, int)

    def __init__(self, database_interface, parent=None):
        super().__init__(parent)

        self.current_simulation = " "
        self.current_index = 0
        self.current_view_orientation = 2
        self.current_scene = 'planning'

        self.graphicsview = View()
        self.graphicsview.request_array.connect(database_interface.request_view_array)
        database_interface.send_view_array.connect(self.graphicsview.cine_film_creation)
        database_interface.send_material_for_viewing.connect(self.view_material)


        self.scenes = {'planning': PlanningScene(),
#                       'running':  RunningScene(),
                       'material': MaterialScene(),
                       'energy imparted': DoseScene(),
                       'dose': DoseScene(front_array='dose'),
                       }
        self.glwidgets = {#'dose': View3D2(arrays=['ctarray','dose'], lut_names=['gist_earth','jet'], dim_scale=False, custom_data_range=[(0, 500), None]),
                          'doseCT': View3D(array='dose', lut_name='jet', dim_scale=True, smoothness=.5, custom_data_range=(0, .15), custom_data_range_is_modifier=True),
#                          'energy_imparted': View3D(array='energy_imparted', lut_name='pet', dim_scale=True),
#                          'planning': View3D(array='ctarray', lut_name='hot_metal_green', dim_scale=False, custom_data_range=(0, 500))}
                          'dens': View3D(array='density', lut_name='gist_earth', dim_scale=True, custom_data_range=(.5, 2.4), smoothness=.5, magic_number = 50),
                          'planning': View3D(array='ctarray', lut_name='gist_earth', dim_scale=False, custom_data_range=(-200, 700), smoothness=.5, magic_number=50)}

        for name, scene in self.scenes.items():
            # connecting scenes to request array slot
            scene.update_index.connect(self.update_index)
            scene.request_array.connect(database_interface.request_view_array)
            database_interface.send_view_array.connect(scene.set_requested_array)

        for name, glwidget in self.glwidgets.items():
            # connecting scenes to request array slot
            glwidget.request_array.connect(database_interface.request_view_array_bytescaled)
            database_interface.send_view_array_bytescaled.connect(glwidget.set_requested_array)


        self.request_array_slice.connect(database_interface.request_view_array_slice)
        database_interface.send_view_array_slice.connect(self.reload_slice)

        self.request_metadata.connect(database_interface.request_simulation_properties)
        database_interface.send_view_sim_propeties.connect(self.set_simulation_properties)
        self.graphicsview.setScene(self.scenes['planning'])

    def set_simulation_editor(self, propertieseditmodel):
        self.scenes['planning'].update_simulation_properties.connect(propertieseditmodel.set_simulation_properties)


    @QtCore.pyqtSlot(str)
    def set_simulation(self, name):
        self.current_simulation = name
#        self.request_metadata.emit(self.current_simulation) # Not needed, propetiesmodel will request metadata

    @QtCore.pyqtSlot(dict)
    def set_simulation_properties(self, data_dict):
        if data_dict.get('name', '') != self.current_simulation:
            return
        for scene in self.scenes.values():
            scene.set_metadata(data_dict)
        if data_dict['is_phantom']:
            for glwid in self.glwidgets.values():
                glwid.set_metadata(data_dict)
        else:
            for glwid in [value for key, value in self.glwidgets.items() if key != 'dens']:
                glwid.set_metadata(data_dict)
        self.selectScene('planning')
        self.update_index(self.current_index)
        self.graphicsview.fitInView(self.graphicsview.sceneRect(), QtCore.Qt.KeepAspectRatio)


    @QtCore.pyqtSlot(int)
    def update_index(self, index):
        self.current_index = index
        for arr_name in self.scenes[self.current_scene].array_names:
            self.request_array_slice.emit(self.current_simulation, arr_name, self.current_index, self.current_view_orientation)

    @QtCore.pyqtSlot(str, str, np.ndarray, int, int)
    def reload_slice(self, name, arr_name, arr, index, orientation):
        if name == self.current_simulation and index == self.current_index:
            self.scenes[self.current_scene].reload_slice(name, arr_name, arr, index, orientation)

    @QtCore.pyqtSlot()
    def selectViewOrientation(self):
        self.current_view_orientation += 1
        self.current_view_orientation %= 3
        for scene in self.scenes.values():
            scene.set_view_orientation(self.current_view_orientation, self.current_index)
        self.update_index(self.current_index)
        self.graphicsview.fitInView(self.graphicsview.sceneRect(), QtCore.Qt.KeepAspectRatio)

    @QtCore.pyqtSlot(str)
    def selectScene(self, scene_name):
        if scene_name in self.scenes:
#            self.graphicsview.show()
            self.current_scene = scene_name
            self.graphicsview.setScene(self.scenes[self.current_scene])
            self.update_index(self.current_index)
            self.graphicsview.fitInView(self.graphicsview.sceneRect(), QtCore.Qt.KeepAspectRatio)
#        else:
#            self.graphicsview.hide()
#        for wids in self.glwidgets.values():
#                wids.hide()
#        if scene_name in self.glwidgets:
##            self.graphicsview.hide()
#            self.glwidgets[scene_name].show()
#            print('showing')


    def view_widget(self):

#        wid = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        wid2 = QtWidgets.QWidget()
        main_layout = QtWidgets.QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        view_layout = QtWidgets.QVBoxLayout()
        view_layout.setContentsMargins(0, 0, 0, 0)

        menu_widget = QtWidgets.QMenuBar(wid2)
        menu_widget.setContentsMargins(0, 0, 0, 0)

        orientation_action = QtGui.QAction('Orientation', menu_widget)
        orientation_action.triggered.connect(self.selectViewOrientation)
        menu_widget.addAction(orientation_action)

        sceneSelect = SceneSelectGroup(wid2)
        for scene_name in self.scenes.keys():
            sceneSelect.addAction(scene_name)
        sceneSelect.scene_selected.connect(self.selectScene)

        main_layout.addWidget(menu_widget)

        main_layout.addWidget(self.graphicsview)

#        for name in self.glwidgets:
#            wid.addWidget(self.glwidgets[name])

        for action in sceneSelect.actions():
            menu_widget.addAction(action)

        cine_action = QtGui.QAction('Cine', menu_widget)
        cine_action.triggered.connect(self.graphicsview.request_cine_film_creation)
        menu_widget.addAction(cine_action)

        wid2.setLayout(main_layout)
        wid = VolumeViewManager(self.glwidgets)
        return wid2, wid



    def make_material_plot_widget(self, material):
        mw = QtWidgets.QMainWindow()
        mw.setWindowTitle(material.name)

        pw = pyqtgraph.PlotWidget(name=material.name, title=material.name, background='w')
        mw.setCentralWidget(pw)
        pw.addLegend()
#        p = win.addPlot(title='plot_title', background='w')
        atts = material.attinuation
        e_ind = atts['energy'] < 500000
        for j, key in enumerate(['total', 'rayleigh', 'photoelectric', 'compton']):
            if key == 'total':
                pen = pyqtgraph.mkPen('k', width=4)
            else:
                pen = pyqtgraph.mkPen(pyqtgraph.intColor(j, hues=4, minHue=240), width=2)
            pw.plot(x=atts['energy'][e_ind], y=atts[key][e_ind], pen=pen, name=key, antialias=True)

        pw.setLabel('left', "Attenuation", units='cm^2/g')
        pw.setLabel('bottom', "Energy", units='eV')
        pw.setLogMode(x=False, y=True)
        pw.setXRange(0, 50000) # 500 keV
        return mw

    @QtCore.pyqtSlot(object)
    def view_material(self, material):
        win = self.make_material_plot_widget(material)
        win.show()
        self.__plot = win


class Scene(QtWidgets.QGraphicsScene):
    update_index = QtCore.pyqtSignal(int)
    request_array = QtCore.pyqtSignal(str, str)
    def __init__(self, parent=None):
        super().__init__(parent)
        self.name = ''
        self.array_names = []
        self.view_orientation = 2
        self.index = 0
        self.shape = np.ones(3, np.int)
        self.spacing = np.ones(3, np.double)
        self.scaling = np.ones(3, np.double)


    def set_metadata(self, sim, index=0):
        self.name = sim.get('name', '')
        self.spacing = sim.get('spacing', np.ones(3, np.double))
        self.shape = sim.get('shape', np.ones(3, np.int))
        self.scaling = sim.get('scaling', np.ones(3, np.double))
        self.index = index % self.shape[self.view_orientation]

    @QtCore.pyqtSlot(str, np.ndarray, str)
    def set_requested_array(self, name, array, array_name):
        pass



    @QtCore.pyqtSlot(str, str, np.ndarray, int, int)
    def reload_slice(self, simulation_name, array_name, arr, index, orientation):
        pass

    def set_view_orientation(self, orientation, index):
        self.view_orientation = orientation
        self.index = index % self.shape[self.view_orientation]

    def wheelEvent(self, ev):
        if ev.delta() > 0:
            self.index += 1
        elif ev.delta() < 0:
            self.index -= 1
        self.index %= self.shape[self.view_orientation]
        self.update_index.emit(self.index)
        ev.accept()


def blendArrayToQImage(f_array, b_array, front_level, back_level,
                       front_lut, back_lut):
    """Convert the 2D numpy array `gray` into a 8-bit QImage with a gray
    colormap.  The first dimension represents the vertical image axis.
    ATTENTION: This QImage carries an attribute `ndimage` with a
    reference to the underlying numpy array that holds the data. On
    Windows, the conversion into a QPixmap does not copy the data, so
    that you have to take care that the QImage does not get garbage
    collected (otherwise PyQt will throw away the wrapper, effectively
    freeing the underlying memory - boom!)."""

    front_array = np.clip(f_array, front_level[0]-front_level[1],
                          front_level[0]+front_level[1])

    front_array -= (front_level[0]-front_level[1])
    front_array *= 255./(front_level[1]*2.)

    back_array = np.clip(b_array, back_level[0]-back_level[1],
                         back_level[0]+back_level[1]).astype(np.float)
    back_array -= (back_level[0]-back_level[1])
    back_array *= 255./(back_level[1]*2.)

    front_array = np.require(front_array, np.uint8, 'C')
    back_array = np.require(back_array, np.uint8, 'C')

    front_h, front_w = front_array.shape
    back_h, back_w = back_array.shape

    front_qim = QtGui.QImage(front_array.data, front_w, front_h, front_w,
                             QtGui.QImage.Format_Indexed8)
    back_qim = QtGui.QImage(back_array.data, back_w, back_h, back_w,
                            QtGui.QImage.Format_Indexed8)
    front_qim.setColorTable(front_lut)
    back_qim.setColorTable(back_lut)

    back_qim = back_qim.convertToFormat(QtGui.QImage.Format_ARGB32_Premultiplied, back_lut)#, flags=QtCore.Qt.DiffuseAlphaDither)

    p = QtGui.QPainter(back_qim)
#    p.setCompositionMode(QtGui.QPainter.CompositionMode_DestinationOut)
#    p.drawImage(QtCore.QRectF(back_qim.rect()), front_qim)
    p.setCompositionMode(QtGui.QPainter.CompositionMode_SourceOver)

    p.drawImage(QtCore.QRectF(back_qim.rect()), front_qim)
#    p.setPen(QtCore.Qt.white)
#    p.drawText(QtCore.QPointF(0, back_w), "({0}, {1})".format(front_level[0], front_level[1]))

    return back_qim


def arrayToQImage(array_un, level, lut):
    """Convert the 2D numpy array `gray` into a 8-bit QImage with a gray
    colormap.  The first dimension represents the vertical image axis.
    ATTENTION: This QImage carries an attribute `ndimage` with a
    reference to the underlying numpy array that holds the data. On
    Windows, the conversion into a QPixmap does not copy the data, so
    that you have to take care that the QImage does not get garbage
    collected (otherwise PyQt will throw away the wrapper, effectively
    freeing the underlying memory - boom!)."""

    WC, WW = level[0], level[1]
    array = np.clip(array_un, WC-WW, WC+WW).astype(np.float)

    array -= (WC - WW)
    array *= 255. / (WW * 2)


#    array = (np.clip(array, WC - 0.5 - (WW-1) / 2, WC - 0.5 + (WW - 1) / 2) -
#             (WC - 0.5 - (WW - 1) / 2)) * 255 / ((WC - 0.5 + (WW - 1) / 2) -
#                                                 (WC - 0.5 - (WW - 1) / 2))
    array = np.require(array, np.uint8, ['C', 'A'])
    h, w = array.shape

    result = QtGui.QImage(array.data, w, h, w, QtGui.QImage.Format_Indexed8)
#    result.ndarray = array
    result.setColorTable(lut)
#    result = result.convertToFormat(QtGui.QImage.Format_ARGB32, lut)
    result.ndarray = array
    return result

def qImageToArray(img, copy=False, transpose=True):
    """
    Convert a QImage into numpy array. The image must have format RGB32, ARGB32, or ARGB32_Premultiplied.
    By default, the image is not copied; changes made to the array will appear in the QImage as well (beware: if
    the QImage is collected before the array, there may be trouble).
    The array will have shape (width, height, (b,g,r,a)).
    """
    fmt = img.format()
    ptr = img.bits()
    ptr.setsize(img.byteCount())
    arr = np.asarray(ptr)
    if img.byteCount() != arr.size * arr.itemsize:
        # Required for Python 2.6, PyQt 4.10
        # If this works on all platforms, then there is no need to use np.asarray..
        arr = np.frombuffer(ptr, np.ubyte, img.byteCount())

    if fmt == img.Format_RGB32:
        arr = arr.reshape(img.height(), img.width(), 3)
    elif fmt == img.Format_ARGB32 or fmt == img.Format_ARGB32_Premultiplied:
        arr = arr.reshape(img.height(), img.width(), 4)

    if copy:
        arr = arr.copy()

    if transpose:
        return arr.transpose((1,0,2))
    else:
        return arr

class NoDataItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.fontMetrics = QtWidgets.qApp.fontMetrics()
        self.msg = "Sorry, no data here yet.\nRun a simulation to compute."
        self.rect = QtCore.QRectF(self.fontMetrics.boundingRect(self.msg))
        self.rect.setHeight(self.rect.height()*2.5)
        self.setFlag(self.ItemIgnoresTransformations, True)
    def boundingRect(self):
#        return  QtCore.QRectF(self.fontMetrics.boundingRect(self.msg))
        return self.rect
#        return QtCore.QRectF(0, 0, 200, 200)

    def paint(self, painter, style, widget=None):
        painter.setPen(QtGui.QPen(QtCore.Qt.white))

        painter.drawText(self.boundingRect(), self.msg)



class BlendImageItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None, overlay_adjust_max_only=True):
        super().__init__(parent)

        self.back_image = np.zeros((8, 8))
        self.back_level = (0, 500)
        self.front_image = np.zeros((8, 8))
        self.front_level = (1000000, 10000)
        if overlay_adjust_max_only:
            self.adjust_front_max_only = True
        else:
            self.adjust_front_max_only = False


        self.back_alpha = 255
        self.front_alpha = 255
        self.back_lut = get_lut('gray', self.back_alpha)
        self.front_lut = get_lut('pet', self.front_alpha)

        self.qimage = None
        self.setAcceptedMouseButtons(QtCore.Qt.RightButton | QtCore.Qt.MiddleButton)

        self.cbar = ColorBarItem(self)
        self.cbar.setParentItem(self)

    def qImage(self):
        if self.qimage is None:
            self.render()
        return self.qimage

    def boundingRect(self):
        return QtCore.QRectF(self.qImage().rect())

    def setImage(self, front_image=None, back_image=None):
        if front_image is not None:
            self.front_image = np.copy(front_image)
        if back_image is not None:
            self.back_image = np.copy(back_image)
        self.prepareGeometryChange()
        self.qimage = None
        self.update(self.boundingRect())

    def setLevels(self, front=None, back=None):
        update = False
        if front is not None:
            self.front_level = front
            self.cbar.set_levels(front)
            update = True
        if back is not None:
            self.back_level = back
            update = True
        if update:
            self.qimage = None
            self.update(self.boundingRect())

    def setLut(self, front_lut=None, back_lut=None, front_alpha=None,
               back_alpha=None):
        update = False
        if front_lut is not None:
            if front_alpha is not None:
                alpha = front_alpha
            else:
                alpha = self.front_alpha
            lut_raw = get_lut(front_lut, alpha)
            self.front_lut = lut_raw
            self.cbar.set_lut(lut_raw)
            update = True
        if back_lut is not None:
            if back_alpha is not None:
                alpha = back_alpha
            else:
                alpha = self.back_alpha
            self.back_lut = get_lut(back_lut, alpha)
            update = True
        if update:
            self.qimage = None
            self.update(self.boundingRect())

    def render(self):
        self.qimage = blendArrayToQImage(self.front_image, self.back_image,
                                         self.front_level, self.back_level,
                                         self.front_lut, self.back_lut)

    def shape(self):
        path = QtGui.QPainterPath()
        path.addEllipse(self.boundingRect())
        return path

    def paint(self, painter, style, widget=None):
        if self.qimage is None:
            self.render()
        painter.drawImage(QtCore.QPointF(0, 0), self.qimage)

    def mousePressEvent(self, event):
        event.accept()

    def mouseMoveEvent(self, event):
        if event.buttons() == QtCore.Qt.RightButton:
            event.accept()
            dp = event.pos()- event.lastPos()
            x, y = self.front_level
            if self.adjust_front_max_only:
                x += dp.x() *.01*abs(x)
                if x < 0:
                    x=0
                y = x
            else:
                x += dp.x()*.01*abs(x)
                y += dp.y()*.01*abs(y)

                if y < 0:
                    y=0
            self.setLevels(front=(x, y))
        elif event.buttons() == QtCore.Qt.MiddleButton:
            event.accept()
            dp = event.pos()- event.lastPos()
            x, y = self.back_level
            x += dp.x()
            y += dp.y()
            if y < 1:
                y=1
            self.setLevels(back=(x, y))

    def setVisible(self, visible):
        self.cbar.setVisible(visible)
        super().setVisible(visible)


class BitImageItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFlag(QtWidgets.QGraphicsItem.ItemSendsGeometryChanges)
        self.image = np.zeros((8, 8), dtype=np.uint8)
        self.prepareGeometryChange()
        self.qimage = None
        self.lut = get_lut('pet')

    def qImage(self):
        if self.qimage is None:
            self.render()
        return self.qimage

    def boundingRect(self):
        return QtCore.QRectF(self.qImage().rect())

    def set_lut(self, lut):
        self.lut = lut

    def setImage(self, image):
        self.image = np.require(image, np.uint8, ['C', 'A'])
        self.prepareGeometryChange()
        self.qimage = None
        self.update(self.boundingRect())

    def render(self):
        h, w = self.image.shape

        self.qimage = QtGui.QImage(self.image.data, w, h, w, QtGui.QImage.Format_Indexed8)
#    result.ndself.image = self.image
        self.qimage.setColorTable(self.lut)
#    result = result.convertToFormat(QtGui.QImage.Format_ARGB32, lut)
        self.qimage.ndarray = self.image

    def shape(self):
        path = QtGui.QPainterPath()
        path.addEllipse(self.boundingRect())
        return path

    def paint(self, painter, style, widget=None):
        painter.drawImage(QtCore.QPointF(0, 0), self.qImage())

class ImageItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFlag(QtWidgets.QGraphicsItem.ItemSendsGeometryChanges)
        self.image = np.zeros((8, 8))
        self.level = (0, 700)
        self.qimage = None
        self.lut = get_lut('gray')
        self.setLevels((0., 1.))

    def qImage(self):
        if self.qimage is None:
            self.render()
        return self.qimage

    def boundingRect(self):
        return QtCore.QRectF(self.qImage().rect())

    def setImage(self, image):
        self.image = image
        self.prepareGeometryChange()
        self.qimage = None
        self.update(self.boundingRect())

    def setLevels(self, level=None):
        if level is None:
            p = self.image.max() - self.image.min()
            level = (p/2., p / 2. * .75)
        self.level = level
        self.qimage = None
        self.update(self.boundingRect())

    def setLut(self, lut):
        self.lut = lut
        self.qimage = None
        self.update(self.boundingRect())

    def render(self):
        self.qimage = arrayToQImage(self.image, self.level,
                                    self.lut)

    def paint(self, painter, style, widget=None):
        painter.drawImage(QtCore.QPointF(self.pos()), self.qImage())

    def mousePressEvent(self, event):
        if event.buttons() == QtCore.Qt.MiddleButton:
            event.accept()

    def mouseMoveEvent(self, event):
        if event.buttons() == QtCore.Qt.MiddleButton:
            event.accept()
#            print(event.pos()- event.lastPos())
            dp = event.pos()- event.lastPos()
            x, y = self.level
            x += dp.x()
            y += dp.y()
            if y < 1:
                y=1
            self.setLevels((x, y))

class AecItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.aec = np.zeros((5, 2))
        self.view_orientation = 2
        self.__shape = (1, 1, 1)
        self.index = 0
        self.__path = None
        self.__path_pos = None

    def set_aec(self, aec, view_orientation, shape):
        self.__shape = shape
        self.view_orientation = view_orientation

        self.aec = aec
        self.aec[:, 0] -= self.aec[:, 0].min()

        self.aec[:, 0] /= self.aec[:, 0].max()

        if self.aec[:, 1].min() == self.aec[:, 1].max():
            self.aec[:, 1] = .5
        else:
            self.aec[:, 1] -= self.aec[:, 1].min()
            self.aec[:, 1] /= self.aec[:, 1].max()

        self.__path = None
        self.__path_pos = None
        self.prepareGeometryChange()

        self.update(self.boundingRect())

    def boundingRect(self):
        shape = tuple(self.__shape[i] for i in [0, 1, 2] if i != self.view_orientation)
        return QtCore.QRectF(0, 0, shape[1], shape[0]*.1)

    def setIndex(self, index):
        self.index = index % self.__shape[self.view_orientation]
        self.__path_pos = None
#        self.prepareGeometryChange()
        self.update(self.boundingRect())

    def setViewOrientation(self, view_orientation):
        self.view_orientation = view_orientation
        self.__path = None
        self.__path_pos = None
        self.prepareGeometryChange()
        self.update(self.boundingRect())

    def aec_path(self):

        if self.__path is None:
            shape = tuple(self.__shape[i] for i in [0, 1, 2] if i != self.view_orientation)
            self.__path = QtGui.QPainterPath()

            x = self.aec[:, 0] * shape[1]
            y = (1. - self.aec[:, 1]) * shape[0] * .1
            self.__path.moveTo(x[0], y[0])
            for i in range(1, self.aec.shape[0]):
                self.__path.lineTo(x[i], y[i])

            self.__path.moveTo(0, 0)
            self.__path.lineTo(0, shape[0]*.1)
            self.__path.lineTo(shape[1], shape[0]*.1)
            self.__path.lineTo(shape[1], 0)
            self.__path.lineTo(0, 0)

        if self.__path_pos is None:
            self.__path_pos = QtGui.QPainterPath()
            if self.view_orientation == 2:
                shape = tuple(self.__shape[i] for i in [0, 1, 2] if i != self.view_orientation)
                self.__path_pos = QtGui.QPainterPath()
                x = self.aec[:, 0] * shape[1]
                y = (1. - self.aec[:, 1]) * shape[0] * .1
                x_c = self.index / self.__shape[2]
                y_c = 1. - np.interp(x_c, self.aec[:, 0], self.aec[:, 1])
                self.__path_pos.addEllipse(QtCore.QPointF(x_c * shape[1], y_c * shape[0] *.1), shape[0]*.005, shape[0]*.01)

        p = QtGui.QPainterPath()
        p.addPath(self.__path)
        p.addPath(self.__path_pos)
        return p

    def paint(self, painter, style, widget=None):
        painter.setPen(QtGui.QPen(QtCore.Qt.white))
        painter.setRenderHint(painter.Antialiasing, True)
        painter.drawPath(self.aec_path())

class ColorBarItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)

        self._array = np.outer(np.linspace(0, 1, 128)[::-1], np.ones(20))
        self.setFlag(self.ItemIgnoresTransformations, True)
        self.fontMetrics = QtWidgets.qApp.fontMetrics()
        self.box_size = self.fontMetrics.boundingRect('A').height()

        self._qim = None

        self._text = ['test', 'lower']
        self._units = 'mGy/100mAs'
        self.setScale(.75)

    def setUnits(self, units):
        self._units = units

    def set_levels(self, level):
        wc, ww = level
        self._text = ["{:.2G} {}".format(wc+ww, self._units), "{:.2G} {}".format(wc-ww, self._units)]
        self.update()

    def boundingRect(self):
        w = max([self._array.shape[1],
                 self.fontMetrics.boundingRect(self._text[0]).width(),
                 self.fontMetrics.boundingRect(self._text[1]).width()])
        h = sum([self._array.shape[0],
                 self.fontMetrics.boundingRect(self._text[0]).height(),
                 self.fontMetrics.boundingRect(self._text[1]).height()])

        return QtCore.QRectF(0, 0, w, h)

    def set_lut(self, lut):
        self._qim = arrayToQImage(self._array, (.5, .5), lut)

    def paint(self, painter, style, widget=None):
        if self._qim is not None:
            im_rect = QtCore.QRectF(0, self.fontMetrics.boundingRect(self._text[0]).height(), self._array.shape[1], self._array.shape[0])
            painter.drawImage(im_rect, self._qim)
            painter.setPen(QtGui.QPen(QtCore.Qt.white))
            painter.drawText(im_rect.topLeft(), self._text[0])
            painter.drawText(self.boundingRect().bottomLeft(), self._text[1])

class PositionBarItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None, callback=None):
        super().__init__(parent)
        self.axes = np.eye(3)
        self.orientation = 2
        self.shape = np.ones(3)
        self.spacing = np.ones(3)
        self.pos = (0, 0)
        self.callback = callback

    def set_data(self, sim):
        self.shape = sim['shape']
        self.spacing = sim['spacing']#*sim['scaling']
        self.set_cosines(sim['image_orientation'])
        self.pos = [sim['start']/self.spacing[2], sim['stop']/self.spacing[2]]
        self.pos.sort()
        self.update()

    def set_orientation(self, orientation):
        self.orientation = orientation
        self.update()

    def set_cosines(self, cosines):
        self.axes[0, :] = cosines[:3]
        self.axes[1, :] = cosines[3:6]
        self.axes[2, :] = np.cross(cosines[:3], cosines[3:6])

    def boundingRect(self):
        x, y = [i for i in range(3) if i != self.orientation]
        return QtCore.QRectF(0, 0, y, x)

    def mousePressEvent(self, event):
        if event.button()==QtCore.Qt.LeftButton and self.callback:
            event.accept()
            pos = event.pos()
            intersect = sum(self.pos) / 2
            if pos.x() < intersect:
                self.pos[0] = pos.x()
                self.callback(start=self.pos[0])
            else:
                self.pos[1] = pos.x()
                self.callback(stop=self.pos[1])

    def paint(self, painter, style, widget=None):
        if self.orientation == 2:
            return
        i, j = [i for i in range(3) if i != self.orientation]

        painter.setPen(QtCore.Qt.white)
        x1 = 0
        x2 = self.shape[i]
        for pos in self.pos:
            y1 = pos
            y2 = pos + self.axes[i, 2] / self.spacing[j] * self.shape[j]
            painter.drawLine(y1, x1, y2, x2)


class PlanningScene(Scene):
    update_simulation_properties = QtCore.pyqtSignal(dict)
    def __init__(self, parent=None):
        super().__init__(parent)
        self.image_item = ImageItem()
        self.image_item_bit = BitImageItem()
        self.position_item = PositionBarItem()
        self.addItem(self.image_item)
        self.addItem(self.image_item_bit)
        self.addItem(self.position_item)
        self.array_names = ['ctarray', 'organ']

        self.aec_item = AecItem()
        self.addItem(self.aec_item)
        self.image_item.setLevels((0, 500))
        self.is_bit_array = False
        self.lut = get_lut('pet')

    def set_metadata(self, sim, index=0):
        super().set_metadata(sim, index)
        self.is_bit_array = sim.get('is_phantom', False)
        self.image_item.setVisible(not self.is_bit_array)
        self.image_item_bit.setVisible(self.is_bit_array)
        self.position_item.set_data(sim)
        if self.is_bit_array:
            self.array_names = ['organ']
            self.request_array.emit(self.name, 'organ_map')
        else:
            self.array_names = ['ctarray']
        self.request_array.emit(self.name, 'exposure_modulation')

        self.updateSceneTransform()

    def position_callback(self, start=None, stop=None):
        update = {}
        if start:
            update['start'] = start
        if stop:
            update['stop'] = stop
        if len(update) > 0:
            update['name'] = self.name
            self.update_simulation_properties.emit(update)

    @QtCore.pyqtSlot(str, np.ndarray, str)
    def set_requested_array(self, name, array, array_name):
        if name != self.name:
            return
        if array_name == 'organ_map':
            organ_max_value = array['organ'].max()
            lut =  [self.lut[i*255 // organ_max_value] for i in range(organ_max_value + 1)]
            self.image_item_bit.set_lut(lut)
        elif array_name == 'exposure_modulation':
            self.aec_item.set_aec(array, self.view_orientation, self.shape)

    def set_view_orientation(self, view_orientation, index):
        self.view_orientation = view_orientation
        self.aec_item.setViewOrientation(view_orientation)
        self.index = index % self.shape[self.view_orientation]
        self.position_item.set_orientation(self.view_orientation)
        self.updateSceneTransform()

    def updateSceneTransform(self):
        sx, sy = [self.spacing[i] for i in range(3) if i != self.view_orientation]
        transform = QtGui.QTransform.fromScale(sy / sx, 1.)
        if self.is_bit_array:
            self.image_item_bit.setTransform(transform)
        else:
            self.image_item.setTransform(transform)

        self.aec_item.setTransform(transform)
        self.aec_item.prepareGeometryChange()
        self.position_item.setTransform(transform)
        shape = tuple(sh for ind, sh in enumerate(self.shape) if ind != self.view_orientation)
        rect = QtCore.QRectF(0, 0, shape[1], shape[0])
        if self.is_bit_array:
            self.aec_item.setPos(self.image_item_bit.mapRectToScene(rect).bottomLeft())
            self.setSceneRect(self.image_item_bit.mapRectToScene(rect).united(self.aec_item.sceneBoundingRect()))
        else:
            self.aec_item.setPos(self.image_item.mapRectToScene(rect).bottomLeft())
            self.setSceneRect(self.image_item.mapRectToScene(rect).united(self.aec_item.sceneBoundingRect()))

    @QtCore.pyqtSlot(str, str, np.ndarray, int, int)
    def reload_slice(self, simulation_name, array_name, arr, index, orientation):
        if simulation_name != self.name:
            return

        if array_name not in self.array_names:
            return
        self.index = index
        if self.is_bit_array:
            self.image_item_bit.setImage(arr)
        else:
            self.image_item.setImage(arr)
        self.aec_item.setIndex(self.index)


#class RunningScene(Scene):
#    def __init__(self, parent=None):
#        super().__init__(parent)
#        self.image_item = ImageItem()
#        self.addItem(self.image_item)
#        self.image_item.setLut(get_lut('hot_iron'))
#        self.nodata_item = NoDataItem()
#        self.addItem(self.nodata_item)
#        self.nodata_item.setVisible(True)
#
#        self.progress_item = self.addText('0 %')
#        self.progress_item.setDefaultTextColor(QtCore.Qt.white)
#
#        self.n_exposures = 1
#        self.array = None
#        self.view_orientation = 1
#
#
#    def set_metadata(self, props, index=0):
#        if props['name'] != self.name:
#            self.setNoData()
#        super().set_metadata(props, index)
#        collimation = props['detector_rows']*props['detector_width']
#        if props['is_spiral']:
#            self.n_exposures = props['exposures'] * (1 + abs(props['start'] - props['stop']) / collimation / props['pitch'])
#        else:
#            self.n_exposures = props['exposures'] * np.ceil(abs(props['start'] - props['stop']) / props['step'])
#            if self.n_exposures < 1:
#                self.n_exposures = 1
#        self.updateSceneTransform()
#
#
#    def setNoData(self):
#        self.array = None
#        self.image_item.setVisible(False)
#        self.progress_item.setVisible(False)
#        self.nodata_item.setVisible(True)
#
#    def defaultLevels(self, array):
#
#        p = array.max() - array.min()
#        return (p/2., p / 2. )
#
#    @QtCore.pyqtSlot(dict, dict)
#    def set_running_data(self, props, arr_dict):
#        self.array = arr_dict.get('energy_imparted', None)
#
#        self.nodata_item.setVisible(False)
#        self.image_item.setVisible(True)
#        self.progress_item.setVisible(True)
#        msg = ""
#        if len(props.get('eta', '')) > 0:
#            if 'start_at_exposure_no' in props:
#                msg += "{} %".format(round(props['start_at_exposure_no'] / self.n_exposures *100, 1))
#            if 'eta' in props:
#                msg += " ETA: {}".format(props['eta'])
#        self.progress_item.setPlainText(msg)
#        if self.array is not None:
#            self.image_item.setLevels(self.defaultLevels(self.array))
#        self.reloadImages()
#        self.updateSceneTransform()
#
#    def set_view_orientation(self, orientation, index):
#        self.view_orientation = orientation
#        self.reloadImages()
#        self.updateSceneTransform()
#
#    def updateSceneTransform(self):
#        sx, sy = [self.spacing[i]*self.scaling[i] for i in range(3) if i != self.view_orientation]
#        transform = QtGui.QTransform.fromScale(sy / sx, 1.)
#        self.image_item.setTransform(transform)
#        if self.nodata_item.isVisible():
#            self.setSceneRect(self.nodata_item.sceneBoundingRect())
#        else:
#            self.setSceneRect(self.image_item.sceneBoundingRect())
#
#    def reloadImages(self):
#        if self.array is not None:
#            self.image_item.setImage(self.array.max(axis=self.view_orientation))
#
#    def wheelEvent(self, ev):
#        pass


class MaterialMapItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.fontMetrics = QtWidgets.qApp.fontMetrics()
        self.box_size = self.fontMetrics.boundingRect('A').height()
        self.rect = QtCore.QRectF(0, 0, self.box_size, self.box_size)
        self.map = []


    def set_map(self, mapping, colors):
        self.map = []

        for ind in range(len(mapping)):
            self.map.append((colors[ind], str(mapping['material_name'][ind], encoding='utf-8')))

        max_str_index = 0
        max_len_str = 0
        for ind, item in enumerate(self.map):
            if len(item[1]) > max_len_str:
                max_len_str = len(item[1])
                max_str_index = ind

        sub_rect = self.fontMetrics.boundingRect(self.map[max_str_index][1])
        self.rect = QtCore.QRectF(0, 0,
                                  self.box_size * 1.25 + sub_rect.width(),
                                  sub_rect.height() * len(self.map) * 2)

    def boundingRect(self):
        return self.rect

    def paint(self, painter, style, widget=None):
        painter.setPen(QtGui.QPen(QtCore.Qt.white))
        painter.setRenderHint(painter.Antialiasing, True)
        h = self.fontMetrics.boundingRect('A').height()
        for ind, value in enumerate(self.map):
            key, item = value
            painter.fillRect(QtCore.QRectF(0, ind*2*h, self.box_size, self.box_size), QtGui.QColor(key))
            painter.drawText(self.box_size * 1.25, ind*2*h + self.box_size, item)
            painter.drawRect(QtCore.QRectF(0, ind*2*h, self.box_size, self.box_size))


class MaterialScene(Scene):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.lut = get_lut('pet')
        self.image_item = BitImageItem()
        self.addItem(self.image_item)
        self.map_item = MaterialMapItem()
        self.addItem(self.map_item)
        self.nodata_item = NoDataItem()
        self.addItem(self.nodata_item)

        self.array_names = ['material']


    def setNoData(self):
        self.nodata_item.setVisible(True)
        self.map_item.setVisible(False)
        self.image_item.setVisible(False)
        self.setSceneRect(self.nodata_item.boundingRect())

    def set_metadata(self, sim, index=0):
        super().set_metadata(sim, index)
        self.setNoData()
        self.index = self.index % self.shape[self.view_orientation]
        self.request_array.emit(self.name, 'material_map')
#        self.updateSceneTransform()

    @QtCore.pyqtSlot(str, np.ndarray, str)
    def set_requested_array(self, name, array, array_name):
        if name != self.name:
            return
        if array_name == 'material_map':
            organ_max_value = array['material'].max()
            lut =  [self.lut[i*255 // organ_max_value] for i in range(organ_max_value+1)]
            self.image_item.set_lut(lut)
            self.map_item.set_map(array, lut)

            self.nodata_item.setVisible(False)
            self.map_item.setVisible(True)
            self.image_item.setVisible(True)
            self.updateSceneTransform()

    def set_view_orientation(self, orientation, index):
        self.view_orientation = orientation
        self.index = index % self.shape[self.view_orientation]
        if not self.nodata_item.isVisible():
            self.updateSceneTransform()


    def updateSceneTransform(self):
        sx, sy = [self.spacing[i]*self.scaling[i] for i in range(3) if i != self.view_orientation]
        transform = QtGui.QTransform.fromScale(sy / sx, 1.)
        self.image_item.setTransform(transform)

        self.map_item.prepareGeometryChange()

#        shape = tuple(sh for ind, sh in enumerate(self.shape) if ind != self.view_orientation)
        shape = tuple(int(self.shape[i] / self.scaling[i]) for i in range(3) if i != self.view_orientation)
        rect = QtCore.QRectF(0, 0, shape[1], shape[0])
        self.map_item.setScale(self.image_item.mapRectToScene(rect).height() / self.map_item.boundingRect().height())
        self.map_item.setPos(self.image_item.mapRectToScene(rect).topRight())
        self.setSceneRect(self.image_item.mapRectToScene(rect).united(self.map_item.sceneBoundingRect()))

    @QtCore.pyqtSlot(str, str, np.ndarray, int, int)
    def reload_slice(self, simulation_name, array_name, arr, index, orientation):
        if simulation_name != self.name:
            return
        if array_name != 'material':
            return
        self.index = index
        self.image_item.setImage(arr)

class DoseScene(Scene):
    def __init__(self, parent=None, front_array='energy_imparted'):
        super().__init__(parent)

        self.image_item = BlendImageItem()
        self.addItem(self.image_item)
#        self.addItem(self.image_item.cbar)
        self.nodata_item = NoDataItem()
        self.addItem(self.nodata_item)
        self.array_names = ['ctarray', 'organ']

        self.front_array_name = front_array
        self.image_item.cbar.setUnits('mGy/100mAs' if front_array == 'dose' else 'eV')

        self.front_array = None

        alpha = 1. -np.exp(-np.linspace(0, 6, 256))
        alpha *= 255./alpha.max()
        self.image_item.setLut(front_lut='jet', front_alpha=alpha.astype(np.int))
        self.setNoData()

    def setNoData(self):
        self.front_array = None
        self.nodata_item.setVisible(True)
        self.image_item.setVisible(False)

    def set_metadata(self, sim, index=0):
        super().set_metadata(sim, index)
        self.setNoData()
#        self.front_array = None
#        self.nodata_item.setVisible(False)
#        self.image_item.setVisible(False)
        self.request_array.emit(self.name, self.front_array_name)
        if sim['is_phantom']:
            self.request_array.emit(self.name, 'organ_map')
        else:
            self.image_item.setLevels(back=(0, 500))

        self.updateSceneTransform()


#    def updateSceneTransform(self):
#        if not self.nodata_item.isVisible():
#            sx, sy = [self.spacing[i] for i in range(3) if i != self.view_orientation]
#            transform = QtGui.QTransform.fromScale(sy / sx, 1.)
#            self.image_item.setTransform(transform)
#
#            shape = tuple(sh for ind, sh in enumerate(self.shape) if ind != self.view_orientation)
#            rect = self.image_item.mapRectToScene(QtCore.QRectF(0, 0, shape[1], shape[0]))
#            c_rect = self.image_item.cbar.boundingRect()
#
#            if rect.height() >= rect.width():
##                self.image_item.cbar.setScale(rect.height() / (10 * c_rect.height()))
#                c_rect = self.image_item.cbar.sceneBoundingRect()
##                self.image_item.cbar.setPos(-c_rect.width(), 0)
#            else:
#
##                self.image_item.cbar.setScale(rect.width()  / (10 * c_rect.height()))
#                c_rect = self.image_item.cbar.sceneBoundingRect()
##                self.image_item.cbar.setPos(0, -c_rect.height())
#
#            self.setSceneRect(rect.united(self.image_item.cbar.sceneBoundingRect()))
#
#
#        else:
#            self.setSceneRect(self.nodata_item.sceneBoundingRect())

    @QtCore.pyqtSlot(str, np.ndarray, str)
    def set_requested_array(self, name, array, array_name):
        if name != self.name:
            return
        if array_name == 'organ_map':
            max_level = array['organ'].max()
            self.image_item.setLevels(back=(max_level/2, max_level/2))
#            self.image_item.setVisible(True)
#            self.nodata_item.setVisible(False)
#            shape = tuple(sh for ind, sh in enumerate(self.shape) if ind != self.view_orientation)
#            rect = self.image_item.mapRectToScene(QtCore.QRectF(0, 0, shape[1], shape[0]))
#            self.setSceneRect(rect)
        elif array_name == self.front_array_name:
            self.front_array = gaussian_filter(array, 1)
            max_level = array.max()/ 4
            min_level = max_level / 4
            self.image_item.setLevels(front=(min_level/2. + max_level/2.,min_level/2. + max_level/2.))
            self.image_item.setVisible(True)
            self.nodata_item.setVisible(False)
            shape = tuple(sh for ind, sh in enumerate(self.shape) if ind != self.view_orientation)
            rect = self.image_item.mapRectToScene(QtCore.QRectF(0, 0, shape[1], shape[0]))
            self.setSceneRect(rect)

    def updateSceneTransform(self):

        sx, sy = [self.spacing[i] for i in range(3) if i != self.view_orientation]
        transform = QtGui.QTransform.fromScale(sy / sx, 1.)
        self.image_item.setTransform(transform)

        if not self.nodata_item.isVisible():
            shape = tuple(sh for ind, sh in enumerate(self.shape) if ind != self.view_orientation)
            rect = self.image_item.mapRectToScene(QtCore.QRectF(0, 0, shape[1], shape[0]))
            self.setSceneRect(rect)
        else:
            self.setSceneRect(self.nodata_item.boundingRect())


    @QtCore.pyqtSlot(str, str, np.ndarray, int, int)
    def reload_slice(self, name, array_name, arr, index, orientation):
        if name != self.name:
            return
        if array_name in self.array_names:
            if self.front_array is not None:
                index_front = (index % self.shape[self.view_orientation]) // self.scaling[self.view_orientation]
                index_front %= self.front_array.shape[self.view_orientation]
                index_front = int(index_front)
                if self.view_orientation == 0:
                    im = np.squeeze(self.front_array[index_front, :, :])[:, :]
                elif self.view_orientation == 1:
                    im = np.squeeze(self.front_array[:, index_front, :])[:, :]
                elif self.view_orientation == 2:
                    im = np.squeeze(self.front_array[:, :, index_front])[:, :]
                self.image_item.setImage(front_image=im, back_image=arr)
        else:
            self.image_item.setImage(back_image=arr)

    @QtCore.pyqtSlot(int, int)
    def set_view_orientation(self, view_orientation, index):
        self.view_orientation = view_orientation % 3
        self.index = index % self.shape[self.view_orientation]
        self.updateSceneTransform()


class RunningScene(QtWidgets.QGraphicsScene):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.image_item = ImageItem()
        self.addItem(self.image_item)
        self.image_item.setLut(get_lut('hot_iron'))
#        self.nodata_item = NoDataItem()
#        self.addItem(self.nodata_item)
#        self.nodata_item.setVisible(True)

        self.progress_item = self.addText('0 %')
        self.progress_item.setDefaultTextColor(QtCore.Qt.white)
        self.progress_item.setFlag(self.progress_item.ItemIgnoresTransformations, True)
        self.progress_item
    def defaultLevels(self, array):

        p = array.max() - array.min()
        return (p/2., p / 2. )

    def set_running_data(self, array, sx, sy, msg):
        self.progress_item.setPlainText(msg)

        self.image_item.setLevels(self.defaultLevels(array))
        self.image_item.setImage(array)
        transform = QtGui.QTransform.fromScale(sy / sx, 1.)
        self.image_item.setTransform(transform)
        self.setSceneRect(self.image_item.sceneBoundingRect())


class RunnerView(QtWidgets.QGraphicsView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.setBackgroundBrush(QtGui.QBrush(QtCore.Qt.black))

        self.setRenderHints(QtGui.QPainter.Antialiasing |
                            QtGui.QPainter.TextAntialiasing)
        self.setScene(RunningScene())
        self.hide()

    def resizeEvent(self, ev):
        super().resizeEvent(ev)
        if self.isVisible():
            self.fitInView(self.sceneRect(), QtCore.Qt.KeepAspectRatio)

    def setScene(self, scene):
        super().setScene(scene)
        self.fitInView(self.sceneRect(), QtCore.Qt.KeepAspectRatio)

    def fitInView(self, *args):
        if self.isVisible():
            super().fitInView(*args)

    @QtCore.pyqtSlot(np.ndarray, float, float, str, bool)
    def set_data(self, array, sx, sy, msg, visible):
        if visible:
            self.show()
            self.scene().set_running_data(array, sx, sy, msg)
            self.fitInView(self.sceneRect(), QtCore.Qt.KeepAspectRatio)
        else:
            self.hide()

class View(QtWidgets.QGraphicsView):
    request_array = QtCore.pyqtSignal(str, str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.setBackgroundBrush(QtGui.QBrush(QtCore.Qt.black))

        self.setRenderHints(QtGui.QPainter.Antialiasing |
                            QtGui.QPainter.TextAntialiasing)

        self.mouse_down_pos = QtCore.QPoint(0, 0)
        self.cine_film_data = {}


    def resizeEvent(self, ev):
        super().resizeEvent(ev)
        if self.isVisible():
            self.fitInView(self.sceneRect(), QtCore.Qt.KeepAspectRatio)

    def setScene(self, scene):
        super().setScene(scene)
        self.fitInView(self.sceneRect(), QtCore.Qt.KeepAspectRatio)

    def fitInView(self, *args):
        if self.isVisible():
            super().fitInView(*args)

    def mouseMoveEvent(self, e):
        if e.buttons() == QtCore.Qt.LeftButton:
            dist = self.mouse_down_pos - e.globalPos()
            if dist.manhattanLength() > QtWidgets.qApplication.startDragDistance():
                e.accept()
                drag = QtGui.QDrag(self)
                # creating mimedata
                qim = self.toQImage()
                md = QtCore.QMimeData()
                md.setImageData(qim)
                drag.setMimeData(md)
                pix = QtGui.QPixmap.fromImage(qim.scaledToWidth(64))
                drag.setPixmap(pix)
                # initialiserer drops
                drag.exec_(QtCore.Qt.CopyAction)
        else:
            return super().mouseMoveEvent(e)
    def mousePressEvent(self, e):
        self.mouse_down_pos = e.globalPos()
        return super().mousePressEvent(e)

    @QtCore.pyqtSlot()
    def request_cine_film_creation(self):
        if not isinstance(self.scene(), Scene):
            return
        self.cine_film_data = {}
        self.cine_film_data['view_orientation'] = self.scene().view_orientation
        self.cine_film_data['array_names'] = self.scene().array_names
        self.cine_film_data['name'] = self.scene().name

        for arr_name in self.cine_film_data['array_names']:
            self.request_array.emit(self.cine_film_data['name'], arr_name)


    @QtCore.pyqtSlot(str, np.ndarray, str)
    def cine_film_creation(self, name, array, array_name):
        if len(self.cine_film_data) == 0:
            return
        if name != self.cine_film_data.get('name', ''):
            return
        if array_name not in self.cine_film_data['array_names']:
            return

        rect = self.toQImage(square=False).rect()
        height, width = rect.height(), rect.width()

        filename = QtWidgets.QFileDialog.getSaveFileName(self,
                                                     "Save cineloop",
                                                     "/cine.mp4",
                                                     "Movie (*.mp4)")
        try:
            writer = FFMPEG_VideoWriter(filename,
                                        (width, height),
                                        15, threads=1)
        except FileNotFoundError:
            logger.warning("FFMPEG executable not found")
            return
        else:
            logger.debug('Writing cine movie')

        for index in range(array.shape[self.cine_film_data['view_orientation']]):
            if self.cine_film_data['view_orientation'] == 1:
                arr_slice = np.ascontiguousarray(np.squeeze(array[:, index, :]))
            elif self.cine_film_data['view_orientation'] == 0:
                arr_slice = np.ascontiguousarray(np.squeeze(array[index, :, :]))
            else:
                arr_slice = np.ascontiguousarray(np.squeeze(array[:,:,index]))

            self.scene().reload_slice(self.cine_film_data['name'],
                                      array_name,
                                      arr_slice,
                                      index,
                                      self.cine_film_data['view_orientation']
                                      )
            QtWidgets.qApp.processEvents(flags=QtCore.QEventLoop.ExcludeUserInputEvents)
            qim = self.toQImage(square=False)#.convertToFormat(QtGui.QImage.Format_RGB888)

            arr = qImageToArray(qim, transpose=False)
#            ptr = qim.bits()
#            ptr.setsize(qim.byteCount())
#            ptr.setsize(width*height*3)

#            arr = np.array(ptr).reshape(height, -1, 3)
#            arr = np.array(ptr).reshape(width, height, 3)
#            arr = np.ascontiguousarray(arr[:,:,:3])
            arr = np.ascontiguousarray(arr[:,:,[2, 1, 0]])
            writer.write_frame(arr)
        writer.close()
        logger.debug('Done writing cine movie')
        self.cine_film_data = {}

    def toQImage(self, scale=True, square=False):
        rect = self.scene().sceneRect()
        h = rect.height()
        h -= h % 4
        w = rect.width()
        w -= w % 4
        b = max([h, w])
        if scale:
            b = max([h, w])
            if b >= 1024:
                s = 1
            else:
                s = int(round(1024/b))
        else:
            s = 1
        if square:
            qim = QtGui.QImage(b, b,
                               QtGui.QImage.Format_ARGB32_Premultiplied)
        else:
            qim = QtGui.QImage(int(w * s), int(h * s),
                               QtGui.QImage.Format_ARGB32_Premultiplied)
        qim.fill(0)
        painter = QtGui.QPainter()
        painter.begin(qim)
        painter.setRenderHints(painter.Antialiasing | painter.TextAntialiasing)
        self.scene().render(painter, source=rect)
        painter.end()
        return qim

class View3Dworker(QtCore.QThread):
    opengl_array = QtCore.pyqtSignal(np.ndarray)

    def __init__(self):
        super().__init__()
        self.data = []
        self.mutex = QtCore.QMutex()

    @QtCore.pyqtSlot(list)
    def generate_tf(self, args):
        self.mutex.lock()
        self.data.append(args)
        self.mutex.unlock()
        self.start()
    def generate_lut(self, array, magic_value=None):
        data=np.arange(256)

        if magic_value is not None:
            corr = (array == int(magic_value)).sum() / (array > 0).sum()
            lut =(np.exp(-(data-magic_value)**2/(15**2))*.1*(1-corr)**2 + .7*np.exp(-(data-255)**2/64**2))*255

#            lut = np.roll(np.kaiser(256, 13), 96) * np.exp(-(255-data)/255) * 128
#            lut[[0, 1, 2]] = 0

        else:
            lut = np.roll(np.kaiser(256, 13), 96) * np.exp(-(255-data)/255) * 50
            lut[[0, 1, 2]] = 0

#            lut = np.roll(np.kaiser(256, 9), 0)*50# * np.exp(-(255-data)/255) * 50
#            lut[[0, 1, 2]] =0


        lut[0]=0
        return lut.astype(np.ubyte)

    def run(self):
        while len(self.data) > 0:
            self.mutex.lock()
            data, magic_value, lut_name, smoothness = self.data.pop(0)
            self.mutex.unlock()

            if smoothness:
                data = gaussian_filter(data, smoothness)
            lut = [np.array(l) for l in get_lut_raw(lut_name)]
            d2 = np.empty(data.shape + (4,), dtype=np.ubyte)

            d2[...,0] = lut[0][data]
            d2[...,1] = lut[1][data]
            d2[...,2] = lut[2][data]

            d2[...,3] = self.generate_lut(data, magic_value)[data]

            self.opengl_array.emit(d2)

class View3D(gl.GLViewWidget):
    request_array = QtCore.pyqtSignal(str, str, float, float, bool)
    request_opengl_array = QtCore.pyqtSignal(list)
    is_showing = QtCore.pyqtSignal(bool)
    def __init__(self, *args, array=None, lut_name=None, dim_scale=False, magic_number=None, custom_data_range=None, smoothness=None, custom_data_range_is_modifier=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.__is_showing = False
        self.setMinimumWidth(100)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        self.name = ''
        if custom_data_range is None:
            self.custom_data_range = (0, 0)
        else:
            self.custom_data_range = custom_data_range
        if custom_data_range_is_modifier:
            self.custom_data_range_is_modifier = True
        else:
            self.custom_data_range_is_modifier = False
        self.magic_number = magic_number
        if smoothness:
            self.smoothness = smoothness
        else:
            self.smoothness = None

        if array:
            self.array_name = array
        if lut_name:
            self.lut_name = lut_name
        else:
            self.lut_name = 'gray'
        self.dim_scaling = dim_scale

        self.shape = np.ones(3, np.int)
        self.spacing = np.ones(3, np.double)
        self.scaling = np.ones(3, np.double)


        self.worker = View3Dworker()
        self.request_opengl_array.connect(self.worker.generate_tf)
        self.worker.opengl_array.connect(self.set_gl_array)

        self.__glitem = None
        self.hide()


    @QtCore.pyqtSlot()
    def save_image_to_clipboard(self):
        w = self.width()
        h = self.height()
        self.repaint()
        pixels = np.zeros((h, w, 4), dtype=np.ubyte)

        glReadPixels(0, 0, w, h, GL_RGBA, GL_UNSIGNED_BYTE, pixels)

        # swap B,R channels for Qt
        tmp = pixels[...,0].copy()
        pixels[...,0] = pixels[...,2]
        pixels[...,2] = tmp
        pixels = pixels[::-1] # flip vertical
        pixels[:, :, 3] = 255 #setting correct alpha for premultiplied RGB
        pixels = np.require(pixels, np.ubyte, ['C', 'A'])
        result = QtGui.QImage(pixels.data, w, h, QtGui.QImage.Format_ARGB32_Premultiplied).convertToFormat(QtGui.QImage.Format_RGB32)
        result.np_data = pixels

        QtWidgets.qApp.clipboard().setImage(result)

    @QtCore.pyqtSlot(float, float)
    def set_custom_data_range(self, minimum, maximum):
        self.custom_data_range = [minimum, maximum]
        if self.__glitem is not None:
            self.removeItem(self.__glitem)
            self.__glitem = None
        self.showing()

    def showing(self):
        if self.__is_showing and self.__glitem == None:
            self.request_array.emit(self.name, self.array_name, self.custom_data_range[0], self.custom_data_range[1], self.custom_data_range_is_modifier)
        self.setVisible(self.__is_showing and self.__glitem != None)
        self.is_showing.emit(self.__is_showing and self.__glitem != None)

    @QtCore.pyqtSlot(bool)
    def set_showing(self, enabled):
        self.__is_showing = enabled
        self.showing()

    def set_metadata(self, sim, index=0):
        if self.name != sim.get('name', ''):
            if self.__glitem is not None:
                self.removeItem(self.__glitem)
                self.__glitem = None

        self.name = sim.get('name', '')
        self.spacing = sim.get('spacing', np.ones(3, np.double))
        self.shape = sim.get('shape', np.ones(3, np.int))
        if self.dim_scaling:
            self.scaling = sim.get('scaling', np.ones(3, np.double))
        else:
            self.scaling = np.ones(3, np.double)

        self.opts['distance'] = np.sum((self.shape*self.spacing/self.scaling)**2)**.5 * 4

#        if self.__glitem is not None:
#            self.removeItem(self.__glitem)
#            self.__glitem = None
        self.showing()
#        self.index = index % self.shape[self.view_orientation]

    @QtCore.pyqtSlot(str, np.ndarray, str)
    def set_requested_array(self, name, data, array_name):
        if name != self.name:
            if self.__glitem is not None:
                self.removeItem(self.__glitem)
                self.__glitem = None
            return
        if array_name != self.array_name:
            return

        if self.__glitem is None:
            self.request_opengl_array.emit([data, self.magic_number, self.lut_name, self.smoothness])


    @QtCore.pyqtSlot(np.ndarray)
    def set_gl_array(self, d2):
        if self.__glitem is not None:
            self.removeItem(self.__glitem)
            self.__glitem = None
        try:
            #self.__glitem = gl.GLVolumeItem(d2, glOptions='additive')
            self.__glitem = gl.GLVolumeItem(d2, glOptions='translucent', smooth=True, sliceDensity=1)
            #self.__glitem = gl.GLVolumeItem(d2, glOptions='opaque')
        except Exception as e:
            print(e)
            return
        S = self.spacing * self.scaling
        scaling =((S/(np.sum(S*S))**.5))
#        scaling = scaling[[0, 2, 1]]
#        self.__glitem.translate(*(sh/2 for sh in d2.shape))
        self.__glitem.scale(*scaling)
#        self.__glitem.translate(*(-self.shape[[2, 0, 1]] / 2 * scaling))
        self.__glitem.translate(*(-self.shape * .5 * scaling / self.scaling))
        try:
            self.addItem(self.__glitem)
        except Exception as e:
            print(e)
            return
        self.showing()




# -*- coding: utf-8 -*-
"""
Created on Fri Aug  7 10:07:58 2015

@author: erlean
"""

import numpy as np
import pydicom as dicom
import os
import itertools
import logging
from scipy.ndimage.interpolation import affine_transform, spline_filter, zoom
from opendxmc.runner.ct_study_runner import ct_runner_validate_simulation
from opendxmc.utils import find_all_files
from opendxmc.database.h5database import Validator
from opendxmc.utils import rebin, rebin_scaling


logger = logging.getLogger('OpenDXMC')


def matrix_scaled(orientation, spacing, spacing_scan):
    iop = np.array(orientation, dtype=np.float).reshape(2, 3).T
    s_norm = np.cross(*iop.T[:])
    R = np.eye(3)
    R[:, :2] = np.fliplr(iop)
    R[:, 2] = s_norm
    M = np.eye(3)
    M[:3, :3] = R*spacing
    return (np.eye(3)/spacing_scan).dot(M)


def matrix(orientation):
    iop = np.array(orientation, dtype=np.float).reshape(2, 3).T
    s_norm = np.cross(*iop.T[:])
    R = np.eye(3)
    R[:, :2] = np.fliplr(iop)
    R[:, 2] = s_norm
    return R


def image_to_world_transform(image_vector, position, orientation, spacing):
    iop = np.array(orientation, dtype=np.float).reshape(2, 3).T
    s_norm = np.cross(*iop.T[:])
    R = np.eye(3)
    R[:, :2] = np.fliplr(iop)
    R[:, 2] = s_norm
    R[:3, :3] *= spacing
    return np.dot(R, image_vector) + position


#def array_from_dicom_list_low_memory(dc_list, scaling):
#    scaling = np.array(scaling)
#    r = int(dc_list[0][0x28, 0x10].value)
#    c = int(dc_list[0][0x28, 0x11].value)
#    n = len(dc_list)
#    shape = np.ceil(np.array([r, c, n]) / scaling).astype(np.int)
#    stride = int(np.ceil(scaling[2]))
#
#    arr = np.empty(shape, dtype=np.int16)
#
#    sub_arr = np.empty((r, c, stride), dtype=np.int16)
#
#    n_ind = 0
#    while (n_ind+1)*stride < n:
#        for i, dc in enumerate(dc_list[n_ind*stride:(n_ind+1)*stride]):
#            sub_arr[:, :, i] = (dc.pixel_array * int(dc[0x28, 0x1053].value) +
#                                int(dc[0x28, 0x1052].value))
##        import pdb
##        pdb.set_trace()
#        arr[:, : , n_ind] = np.squeeze(zoom(sub_arr, 1./scaling))
#
#        n_ind += 1
#
#    if n_ind*stride < n:
#
#        for i, dc in enumerate(dc_list[n_ind*stride:]):
#            sub_arr[:, :, i] = (dc.pixel_array * int(dc[0x28, 0x1053].value) +
#                                int(dc[0x28, 0x1052].value))
#        for j in range(i, stride):
#            sub_arr[:, :, j] = sub_arr[:, :, i]
#
#        arr[:, : , n_ind] = np.squeeze(zoom(sub_arr, 1./scaling))
#
#    return arr


def array_from_dicom_list(dc_list, scaling):
    r = int(dc_list[0][0x28, 0x10].value)
    c = int(dc_list[0][0x28, 0x11].value)
    n = len(dc_list)

    arr = np.empty((r, c, scaling[2]), dtype=np.int16)
    for i in range(min((n, scaling[2]))):
        dc = dc_list[i]
        arr[:, :, i] = (dc.pixel_array * int(dc[0x28, 0x1053].value) +
                        int(dc[0x28, 0x1052].value))
    if n <= scaling[2]:
        return arr[:, :, :n].mean(axis=2)

    arr_r = rebin_scaling(arr, scaling)
    teller = 0
    for i in range(scaling[2], n):
        dc = dc_list[i]
        arr[:, :, teller] = (dc.pixel_array * int(dc[0x28, 0x1053].value) +
                        int(dc[0x28, 0x1052].value))
        teller += 1
        if teller >= scaling[2]:
            teller = 0
            arr_r = np.concatenate((arr_r, rebin_scaling(arr, scaling)), axis=2)

    return arr_r


def aec_from_dicom_list(dc_list, iop, spacing):
    n_im = len(dc_list)
    exp = np.empty((n_im, 2), dtype=np.float)
    pos = np.zeros(3)

    for i, dc in enumerate(dc_list):
        exp[i, 1] = float(dc[0x18, 0x1152].value)
        exp[i, 0] = image_to_world_transform(np.array([0, 0, i]), pos, iop, spacing)[2]
    exp[:,0] /= 10.0 # for cm units
    return exp


def dc_slice_indicator(dc):
    """Returns a number indicating slce z position"""
    pos = np.array(dc[0x20, 0x32].value)
    iop = np.array(dc[0x20, 0x37].value).reshape((2, 3)).T
    return np.inner(pos, np.cross(*iop.T[:]))


def z_stop_estimator(iop, spacing, shape):
    choices = []
    pos = np.zeros(3)
    for i in [0, shape[0]]:
        for j in [0, shape[1]]:
            for k in [0, shape[2]]:
                choices.append(image_to_world_transform(np.array([i, j, k]), pos, iop, spacing)[2])
    return min(choices), max(choices)



def import_ct_series(paths, import_scaling=(2, 2, 2)):

    import_scaling=np.rint(np.array(import_scaling)).astype(np.int)
    import_scaling[import_scaling < 1] = 1

    series = {}
    for p in find_all_files(paths):
        try:
            dc = dicom.read_file(p)
        except dicom.filereader.InvalidDicomError:
            logger.debug("Not imported: {} -Invalid dicom file".format(p))
        else:
            # test for ct image
            try:
                sop_class_uid = str(dc[0x8, 0x16].value)
            except KeyError:
                logger.debug('Not imported: {} -No SOP UID for file.'.format(p))
                continue
            
            if sop_class_uid == "1.2.840.10008.5.1.4.1.1.2":
                axial_image = str(dc[0x8, 0x8].value[2]).lower()
                if axial_image == 'axial':
                    series_uid = str(dc[0x20, 0xe].value)
                    if series_uid in series:
                        series[series_uid].append(p)
                    else:
                        series[series_uid] = [p]
                    logger.debug("Imported {}".format(p))
                else:
                    logger.debug("Not imported: {} -Image not axial, possible a scout".format(p))
            else:
                logger.debug("Not imported: {0} -Not a CT image:{1}".format(p, sop_class_uid))

    logger.info("Importing {0} CT series with total {1} images".format(
                len(series), sum([len(x) for x in series.values()])))

    for name, series_paths in series.items():

        if len(series_paths) < 2:
            logger.info('Image series {} is skipped since it contains less than 2 images.'.format(name))
            continue
        logger.debug('Setting up data for simulation {}'.format(name))
        dc_list = [dicom.read_file(p) for p in series_paths]
#        dc_list.sort(key=lambda x: x[0x20, 0x13].value)
        dc = dc_list[0]
        dc_list.sort(key=lambda x: dc_slice_indicator(x))

        spacing = np.empty(3, dtype=np.float)
        spacing[:2] = np.array(dc[0x28, 0x30].value)
        spacing[2] = np.sum((np.array(dc_list[1][0x20, 0x32].value) -
                            np.array(dc_list[0][0x20, 0x32].value))**2)**.5

        #Creating transforrmation matrix
        patient = Validator()
        patient.name = name
        patient.import_scaling = import_scaling

        patient.exposure_modulation = aec_from_dicom_list(dc_list,
                                                          np.array(dc[0x20, 0x37].value),
                                                          spacing)

#        try:
        patient.ctarray = array_from_dicom_list(dc_list, import_scaling)
#        except MemoryError:
#            logger.warning('Memory error when importing {}. Attempting low memory import method.'.format(name))
#            patient.ctarray = array_from_dicom_list_low_memory(dc_list, import_scaling)
        patient.shape = np.array(patient.ctarray.shape, dtype=np.int)
        patient.spacing = spacing / 10. * np.array(import_scaling)
        patient.image_position = np.array(dc[0x20, 0x32].value) / 10.
        patient.image_orientation = np.array(dc[0x20, 0x37].value)



        try:
            patient.data_center = np.array(dc[0x18, 0x9313].value) / 10. - patient.image_position
        except KeyError:
            patient.data_center = image_to_world_transform(np.array(patient.shape) / 2.,
                                                           np.zeroes(3),
                                                           patient.image_orientation,
                                                           patient.spacing)

        tag_key = {'pitch': (0x18, 0x9311),
                   'scan_fov': (0x18, 0x90),
                   'sdd': (0x18, 0x1110),
                   'aquired_kV': (0x18, 0x60),
                   'detector_width': (0x18, 0x9306),
                   'region': (0x18, 0x15)
                   }
        units_in_mm = ['scan_fov',
                       'sdd',
                       'detector_width',
                       ]
        for key, tag in tag_key.items():
            try:
                dc[tag].value
            except KeyError:
                pass
            else:
                if key in units_in_mm:
                    setattr(patient, key, float(dc[tag].value) / 10.)
                else:
                    setattr(patient, key, dc[tag].value)
        if patient.aquired_kV > 0:
            patient.kV = patient.aquired_kV
        # correction for sdd
        try:
            patient.sdd = dc[0x18, 0x1111].value * 2 / 10.
        except KeyError:
            pass

        patient.is_spiral = patient.pitch != 0.
        if not patient.is_spiral:
            patient.step = abs(dc_list[0][0x20, 0x32].value[2] -
                               dc_list[1][0x20, 0x32].value[2])

        try:
            total_collimation = dc[0x18, 0x9307].value / 10.
        except KeyError:
            pass
        else:
            patient.detector_rows = (total_collimation /
                                      patient.detector_width)
        try:
            exposure = float(dc[0x18, 0x1152].value)
            ctdi = float(dc[0x18, 0x9345].value)
        except KeyError:
            pass
        else:
            if exposure > 0:
                if patient.is_spiral:
                    patient.ctdi_w100 = ctdi / exposure / patient.pitch * 100.
                else:
                    patient.ctdi_w100 = ctdi / exposure * 100.

        start, stop = z_stop_estimator(patient.image_orientation, patient.spacing, patient.ctarray.shape)
        patient.start_scan = start
        patient.stop_scan = stop
        patient.start = start
        patient.stop = stop


        yield patient.get_data()




# -*- coding: utf-8 -*-
"""
Created on Mon Dec 28 22:26:24 2015

@author: ander
"""

import numpy as np
import scipy.interpolate
import itertools
from opendxmc.tube.tungsten import specter, attinuation
import logging
logger = logging.getLogger('OpenDXMC')


def half_shuffle(arr):
    """
    Shuffles an array in a predictable manner
    """
    assert len(arr.shape) == 1
    
    h = arr.shape[0] // 2
    shuf = np.empty_like(arr)
    shuf[::2] = arr[h:]
    shuf[1::2] = arr[:h][::-1]
    return shuf


def bowtie_path_lenght(angles, radius, distance):
    angle_max = np.arcsin(radius / (radius+distance))
    angle_max_ind = np.abs(angles ) > angle_max

    theta = np.arctan(distance*np.tan(angles)/radius)
    c = radius*(1-np.cos(theta))
    c[angle_max_ind] = radius
    return c / np.cos(angles)


def ct_source_space(simulation, exposure_modulation=None):
    if simulation.get('use_tube_B'):
        wA = simulation.get('tube_weight_A')
        wB = simulation.get('tube_weight_A')
        wS = wA + wB
        
        return itertools.chain(
            ct_source_space_single(simulation, exposure_modulation, tube='A', weight=2*wA/wS),
            ct_source_space_single(simulation, exposure_modulation, tube='B', weight=2*wB/wS),
        )
    else :
        return ct_source_space_single(simulation, exposure_modulation, tube='A')
    

def ct_source_space_single(simulation, exposure_modulation=None, tube='A', weight=1.0):

    
    arglist = ['scan_fov', 'sdd']
    kwarglist = ['start', 'stop', 'exposures', 'histories',
                 'start_at_exposure_no',
                 'bowtie_distance', 'bowtie_radius']

    args = [simulation.get(a) for a in arglist]
    args.append(simulation.get('detector_rows') * simulation.get('detector_width'))
#    import pdb;pdb.set_trace()
    if simulation['use_AEC']:
        kwargs = {'exposure_modulation': exposure_modulation}
    else:
        kwargs = dict()
    for a in kwarglist:
        kwargs[a] = simulation.get(a)
    kwargs['rotation_center'] = simulation.get('data_center')
    kwargs['rotation_plane_cosines'] = simulation.get('image_orientation')
    kwargs['tube_start_angle'] = simulation.get('tube_start_angle_'+tube)
    kwargs['weight'] = weight

    kwargs['xcare'] = simulation.get('xcare')

    if simulation.get('is_spiral'):
        kwargs['pitch'] = simulation.get('pitch')
        phase_func = ct_spiral
    else:
        kwargs['step'] = simulation.get('step')
        phase_func = ct_seq

    s = specter(simulation.get('kV_'+tube), angle_deg=simulation['anode_angle'], filtration_materials='Al',
                filtration_mm=simulation.get('al_filtration'))
    kwargs['energy_specter'] = s

    return phase_func(*args, **kwargs)

def world_image_matrix(orientation):
    iop = np.array(orientation, dtype=np.float).reshape(2, 3).T
    s_norm = np.cross(*iop.T[:])
    R = np.eye(3)
    R[:, :2] = np.fliplr(iop)
    R[:, 2] = s_norm
    return np.linalg.inv(R)

def rotation_z_matrix(alpha):
    return np.array([[np.cos(alpha), - np.sin(alpha), 0],
                     [np.sin(alpha), np.cos(alpha), 0],
                     [0, 0, 1]], dtype=np.double)

def ct_spiral(scan_fov, sdd, total_collimation, pitch=1,
              start=0, stop=1, exposures=100, histories=1,
              tube_start_angle=0.,
              energy=70000., energy_specter=None,
              rotation_center=None,
              rotation_plane_cosines=None,
              exposure_modulation=None, start_at_exposure_no=0,
              bowtie_radius=1, bowtie_distance=0,
              weight=1.0, xcare=False):
    """Generate CT phase space, return a iterator.

    INPUT:
        scan_fov : float
            scanner field of view in cm
        sdd : float
            source detector distance in cm
        total_collimation : float
            detector collimation or width in cm
        pitch : float
            pitch value of the spiral
        start : float
            start z position for spiral in cm
        stop : float
            stop z position for spiral in cm
        exposures : int
            number of exposures per rotation
        histories : int
            number of photon histories per exposure
        energy : float
            monochromatic photon energy in eV (ignored if energy_specter
            is applied)
        energy_specter : [(N,), (N,)]
            [ndarray(energy), ndarray(intensity)] list/tuple of
            two ndarrays of lenght one, with energy and intensity of specter
        modulation_xy : [(N,), (N,)] (NOT IMPLEMENTED)
            tube current XY modulation, list/tuple of
            (ndarray(position), ndarray(scale_factors))
        exposure_modulation : [(N,), (N,)]
            tube current modulation z axis, list/tuple of
            (ndarray(position), ndarray(scale_factors))
        start_at_exposure_no: int
            Starting at this exposure number, used for resuming a simulation
        weight: float
            weight of phasespace 
        xcare: bool
            Use XCare
    OUTPUT:
        Iterator returning ndarrays of shape (8, batch_size),
        one row is equal to photon (start_x, start_y, star_z, direction_x,
        direction_y, direction_z, energy, weight)
    """
    logger.debug('Generating CT spiral phase space')
    if rotation_center is None:
        rotation_center = np.zeros(3, dtype=np.double)
    rotation_center[2] = 0  # we start spiral at start not center
    if rotation_plane_cosines is None:
        rotation_plane_cosines = np.array([1, 0, 0, 0, 1, 0], dtype=np.double)
#    print('CT phase space start', start, stop)
    # total number of exposures + one total rotation
    exposures = int(exposures)
    e = int((abs(start - stop) / (pitch * total_collimation) + 1) * exposures)

    if start < stop:
        d_col = total_collimation / 2.
    else:
        d_col = -total_collimation / 2.
    # positions along z for each exposure
    t = np.linspace(start-d_col, stop + d_col, e)
#    # we shuffle the positions to take generate conservative ETA estimates
    t = half_shuffle(t)

    # angle for each z position , i.e the x, y coordinates
    ang = np.mod(t / (pitch * total_collimation) * np.pi * 2. + np.deg2rad(tube_start_angle), 2*np.pi)
    # rotation matrix along z-axis for an angle x

    if energy_specter is None:
        energy_specter = [np.array([energy], dtype=np.double),
                          np.array([1.0], dtype=np.double)]
    energy_specter = (energy_specter[0],
                      energy_specter[1] / energy_specter[1].sum())

    specter_cpd = np.cumsum(energy_specter[1]).astype('float64')
    specter_cpd /= specter_cpd.max()

    specter_energy = energy_specter[0].astype('float64')

#    if modulation_xy is None:
#        mod_xy = lambda x: 1.0
#    else:
#        mod_xy = scipy.interpolate.interp1d(modulation_xy[0], modulation_xy[1],
#                                            copy=False, bounds_error=False,
#                                            fill_value=1.0)
    
    if exposure_modulation is None:
        mod_z = lambda x: weight
    else:
        if np.abs(np.mean(exposure_modulation[:, 1])) > 0.000001:
            modulator_array = exposure_modulation[:, 1]
            modulator_array /= np.mean(modulator_array)
            modulator_array *= weight
            mod_z = scipy.interpolate.interp1d(exposure_modulation[:, 0],
                                               modulator_array,
                                               copy=True, bounds_error=False,
                                               fill_value=weight, kind='nearest')
        else:
            mod_z = lambda x: weight
    fov_arr=np.array([scan_fov], dtype='float64')
    collimation_arr=np.array([total_collimation], dtype='float64')
    rot_fan_angle = np.array([np.arctan(fov_arr[0]/sdd) * 2],dtype='float64')
    scan_fan_angle = np.array([np.arctan(collimation_arr[0] *.5 / sdd) * 2], dtype='float64')

    bowtie_angle = np.linspace(-rot_fan_angle[0]/2, rot_fan_angle[0]/2, 101, dtype='float64')
    bowtie_lenghts= bowtie_path_lenght(bowtie_angle, bowtie_radius, bowtie_distance)
    bowtie_weights = np.empty_like(bowtie_angle, dtype='float64')
    bowtie_att = attinuation(specter_energy/1000, name='aluminum', density=True).astype('float64')
    for i in range(bowtie_lenghts.shape[0]):
        bowtie_weights[i] = np.sum(energy_specter[1]*np.exp(-bowtie_att*bowtie_lenghts[i]))


    n_bowtie = np.array(bowtie_weights.shape, dtype='int')
    n_specter = np.array(specter_energy.shape, dtype='int')

    if xcare:
        xcare_ang = np.deg2rad(np.array([-60, -40, 40, 60], dtype='float64')+90.0)
        xcare_weight = np.array([1.25, 0.25, 0.25, 1.25], dtype='float64')
        xcare_mod = scipy.interpolate.interp1d(xcare_ang,
                                               xcare_weight,
                                               copy=True, bounds_error=False,
                                               fill_value=(xcare_weight[0], xcare_weight[-1]), kind='linear')
    else:
        xcare_mod = lambda ang: 1.0
        
    M = world_image_matrix(rotation_plane_cosines)
    rotation_center_image = np.dot(M, rotation_center[[1, 0, 2]])
    for i in range(start_at_exposure_no, e):
        R = np.dot(M, rotation_z_matrix(ang[i]))

        position = np.dot(R, np.array([-sdd/2., 0, t[i]], dtype='float64')) + rotation_center_image
        direction = np.dot(R, np.array([1., 0, 0], dtype='float64'))
        scan_axis = np.dot(R, np.array([0, 0, 1], dtype='float64'))
        exposure_weight = np.array([mod_z(t[i])], dtype='float64')
        exposure_weight *= xcare_mod(ang[i])
        ret = (position, direction, scan_axis,
               scan_fan_angle,
               rot_fan_angle,
               exposure_weight,
               specter_cpd, specter_energy, n_specter,
               bowtie_weights, bowtie_angle, n_bowtie)
#        ret = (position, direction, scan_axis,
#               np.array([sdd], dtype='float64'),
#               np.array([scan_fov], dtype='float64'),
#               np.array([total_collimation], dtype='float64'),
#               np.array([mod_z(t[i])], dtype='float64'),
#               specter_cpd.astype('float64'), specter_energy.astype('float64'))
#        print('Weight: {}'.format(ret[5]), exposure_modulation[0, 0], t[i] ,exposure_modulation[-1, 0] )
        yield ret, i, e




def ct_seq(scan_fov, sdd, total_collimation, step=1,
              start=0, stop=1, exposures=100, histories=1,
              tube_start_angle=0.,
              energy=70000., energy_specter=None,
              rotation_center=None,
              rotation_plane_cosines = None,
              bowtie_radius=1, bowtie_distance=0,
              exposure_modulation=None, start_at_exposure_no=0,
              weight=1.0, xcare=False):
    """Generate CT phase space, return a iterator.

    INPUT:
        scan_fov : float
            scanner field of view in cm
        sdd : float
            source detector distance in cm
        total_collimation : float
            detector collimation or width in cm
        pitch : float
            pitch value of the spiral
        start : float
            start z position for spiral in cm
        stop : float
            stop z position for spiral in cm
        exposures : int
            number of exposures per rotation
        histories : int
            number of photon histories per exposure
        energy : float
            monochromatic photon energy in eV (ignored if energy_specter
            is applied)
        energy_specter : [(N,), (N,)]
            [ndarray(energy), ndarray(intensity)] list/tuple of
            two ndarrays of lenght one, with energy and intensity of specter
        modulation_xy : [(N,), (N,)] (NOT IMPLEMENTED)
            tube current XY modulation, list/tuple of
            (ndarray(position), ndarray(scale_factors))
        exposure_modulation : [(N,), (N,)]
            tube current modulation z axis, list/tuple of
            (ndarray(position), ndarray(scale_factors))
        start_at_exposure_no: int
            Starting at this exposure number, used for resuming a simulation
    OUTPUT:
        Iterator returning ndarrays of shape (8, batch_size),
        one row is equal to photon (start_x, start_y, star_z, direction_x,
        direction_y, direction_z, energy, weight)
    """
    logger.debug('Generating CT seq phase space')
    if rotation_center is None:
        rotation_center = np.zeros(3, dtype=np.double)
    rotation_center[2] = 0  # we start spiral at start not center
    if rotation_plane_cosines is None:
        rotation_plane_cosines = np.array([1, 0, 0, 0, 1, 0], dtype=np.double)

    # total number of exposures + one total rotation
    exposures = int(exposures)
    N = int(np.ceil(abs(start - stop) / step))
    if N <= 0:
        N = 1
    e = N * exposures

    d_col = total_collimation / 2.

    if start > stop:
        start, stop = stop, start
    # positions along z for each exposure
    t = np.empty((e,))
    ang = np.empty((e,))
    for i in range(N):
        t[exposures*i: exposures*(i+1)] = start + step * i
        ang[exposures*i: exposures*(i+1)] = np.linspace(0, 2*np.pi, exposures) + np.deg2rad(tube_start_angle)

    t = half_shuffle(t)
    ang = half_shuffle(ang)

    if energy_specter is None:
        energy_specter = [np.array([energy], dtype=np.double),
                          np.array([1.0], dtype=np.double)]
    energy_specter = (energy_specter[0],
                      energy_specter[1] / energy_specter[1].sum())
    specter_cpd = np.cumsum(energy_specter[1]).astype('float64')
    specter_cpd /= specter_cpd.max()
    specter_energy = energy_specter[0].astype('float64')

#    if modulation_xy is None:
#        mod_xy = lambda x: 1.0
#    else:
#        mod_xy = scipy.interpolate.interp1d(modulation_xy[0], modulation_xy[1],
#                                            copy=False, bounds_error=False,
#                                            fill_value=1.0)

    fov_arr=np.array([scan_fov], dtype='float64')
    collimation_arr=np.array([total_collimation], dtype='float64')
    rot_fan_angle = np.array([np.arctan(fov_arr[0]/sdd) * 2],dtype='float64')
    scan_fan_angle = np.array([np.arctan(collimation_arr[0] *.5 / sdd) * 2], dtype='float64')

    bowtie_angle = np.linspace(-rot_fan_angle[0]/2, rot_fan_angle[0]/2, 101, dtype='float64')
    bowtie_lenghts= bowtie_path_lenght(bowtie_angle, bowtie_radius, bowtie_distance)
    bowtie_weights = np.empty_like(bowtie_angle, dtype='float64')
    bowtie_att = attinuation(specter_energy/1000, name='aluminum', density=True).astype('float64')
    for i in range(bowtie_lenghts.shape[0]):
        bowtie_weights[i] = np.sum(energy_specter[1]*np.exp(-bowtie_att*bowtie_lenghts[i]))

    n_bowtie = np.array(bowtie_weights.shape, dtype='int')
    n_specter = np.array(specter_energy.shape, dtype='int')

    if xcare:
        xcare_ang = np.deg2rad(np.array([-60, -40, 40, 60], dtype='float64') +90.0)
        xcare_weight = np.array([1.25, 0.25, 0.25, 1.25], dtype='float64')
        xcare_mod = scipy.interpolate.interp1d(xcare_ang,
                                               xcare_weight,
                                               copy=True, bounds_error=False,
                                               fill_value=(xcare_weight[0], xcare_weight[-1]), kind='linear')
    else:
        xcare_mod = lambda ang: 1.0

    if exposure_modulation is None:
        mod_z = lambda x: weight
    else:
        if np.abs(np.mean(exposure_modulation[:, 1])) > 0.000001:
            modulator_array = exposure_modulation[:, 1]
            modulator_array /= np.mean(modulator_array)
            modulator_array *= weight
            mod_z = scipy.interpolate.interp1d(exposure_modulation[:, 0],
                                               modulator_array,
                                               copy=True, bounds_error=False,
                                               fill_value=weight, kind='nearest')
        else:
            mod_z = lambda x: weight

    M = world_image_matrix(rotation_plane_cosines)
    rotation_center_image = np.dot(M, rotation_center[[1, 0, 2]])
    for i in range(start_at_exposure_no, e):
        R = np.dot(M, rotation_z_matrix(ang[i]))

        position = np.dot(R, np.array([-sdd/2., 0, t[i]], dtype='float64')) + rotation_center_image
        direction = np.dot(R, np.array([1., 0, 0], dtype='float64'))
        scan_axis = np.dot(R, np.array([0, 0, 1], dtype='float64'))
        
        exposure_weight = np.array([mod_z(t[i])], dtype='float64')
        exposure_weight *= xcare_mod(ang[i])
        ret = (position, direction, scan_axis,
               scan_fan_angle,
               rot_fan_angle,
               exposure_weight,
               specter_cpd, specter_energy, n_specter,
               bowtie_weights, bowtie_angle, n_bowtie)

#        ret = (position, direction, scan_axis,
#               np.array([sdd], dtype='float64'),
#               np.array([scan_fov], dtype='float64'),
#               np.array([total_collimation], dtype='float64'),
#               np.array([mod_z(t[i])], dtype='float64'),
#               specter_cpd.astype('float64'), specter_energy.astype('float64'))
        yield ret, i, e




#
from opendxmc.runner.ct_study_runner import ct_runner
from opendxmc.runner.ct_study_runner import obtain_ctdiair_conversion_factor
from opendxmc.runner.ct_study_runner import obtain_ctdiw_conversion_factor
from opendxmc.runner.ct_study_runner import ct_runner_validate_simulation
from opendxmc.runner.ct_study_runner import generate_attinuation_lut


# -*- coding: utf-8 -*-
"""
Created on Mon Jul 27 11:42:37 2015

@author: erlean
"""
import sys
import os
from PyQt5 import QtGui, QtCore, QtWidgets
from opendxmc.app.view import ViewController, RunnerView
from opendxmc.app.model import DatabaseInterface, ListView, ListModel, RunManager, Importer, ImportScalingEdit, PropertiesEditWidget, OrganDoseModel, OrganDoseView
import opendxmc
import logging

logger = logging.getLogger('OpenDXMC')
logger.setLevel(10)
LOG_FORMAT = ("[%(asctime)s %(name)s %(levelname)s]  -  %(message)s  -  in method %(funcName)s line:"
    "%(lineno)d filename: %(filename)s")


class LogHandler(QtCore.QObject, logging.Handler):
    message = QtCore.pyqtSignal(str)

    def __init__(self, parent=None):
        QtCore.QObject.__init__(self, parent)
        logging.Handler.__init__(self)

        self.setFormatter(logging.Formatter(LOG_FORMAT, '%H:%M'))
#        self.log_formater = logging.Formatter()

    def emit(self, log_record):
        self.message.emit(self.format(log_record) + os.linesep)


class LogWidget(QtWidgets.QTextEdit):
    closed = QtCore.pyqtSignal(bool)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setReadOnly(True)

    def closeEvent(self, event):
        self.closed.emit(False)
        super().closeEvent(event)
    
    @QtCore.pyqtSlot(str)
    def insertPlainText(self, txt):
        self.moveCursor(QtGui.QTextCursor.End)
        super().insertPlainText(txt)
        self.ensureCursorVisible()


class SelectDatabaseWidget(QtWidgets.QWidget):
    set_database_path = QtCore.pyqtSignal(QtCore.QUrl)
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.setContentsMargins(0, 0, 0, 0)
  
        help_msg = "Location of database file"        
        self.setToolTip(help_msg)        
        
        settings = QtCore.QSettings('OpenDXMC', 'gui')
        if settings.contains('database/path'):
            self.path = QtCore.QUrl.fromLocalFile(settings.value('database/path'))
        else:
            self.path = QtCore.QUrl.fromLocalFile(os.path.join(os.path.dirname(sys.argv[0]), 'database.h5'))
            settings.setValue('database/path', self.path.toLocalFile())
        
        self.applybutton = QtWidgets.QToolButton()
        self.applybutton.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        self.applybutton.setText('Apply')
        self.applybutton.setContentsMargins(0, 0, 0, 0)
        
        self.txtedit = QtWidgets.QLineEdit()
        self.txtedit.setToolTip(help_msg)
        self.txtedit.setContentsMargins(0, 0, 0, 0)
        self.txtedit.setText(self.path.toLocalFile())
        completer = QtWidgets.QCompleter(self)
        self.model = QtWidgets.QFileSystemModel(completer)
        self.model.setRootPath(os.path.dirname(self.path.toLocalFile()))
        completer.setModel(self.model)
        self.txtedit.setCompleter(completer)
        
        self.layout().addWidget(self.txtedit, 10)
        self.layout().addWidget(self.applybutton, 1)
    
        self.applybutton.clicked.connect(self.apply)    
        self.applybutton.setToolTip(help_msg)
#        self.setMaximumHeight(max([self.txtedit.minimumHeight(), self.applybutton.minimumHeight()]))
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
    
    @QtCore.pyqtSlot()    
    def apply(self):
        self.path = QtCore.QUrl.fromLocalFile(self.txtedit.text())
        self.set_database_path.emit(self.path)
   
    @QtCore.pyqtSlot(QtCore.QUrl)    
    def validate_apply(self, url):
        self.path = url
        self.txtedit.setText(self.path.toLocalFile())
        settings = QtCore.QSettings('OpenDXMC', 'gui')
        settings.setValue('database/path', self.path.toLocalFile())
        self.model.setRootPath(self.path.toLocalFile())
    
    @QtCore.pyqtSlot(bool)    
    def locked(self, val):
        self.setDisabled(val)
        self.applybutton.setDisabled(val)
    
   
class StatusBarButton(QtWidgets.QPushButton):
    def __init__(self, *args):
        super().__init__(*args)
        self.setFlat(True)
        self.setCheckable(True)


class BusyWidget(QtWidgets.QWidget):
    def __init__(self, parent=None, tooltip=''):
        super().__init__(parent)
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.progress)
        self.timer.timeout.connect(self.update)
        self.progress = 0
        self.pen = QtGui.QPen(QtGui.QBrush(QtCore.Qt.white), 50.,
                              cap=QtCore.Qt.RoundCap)
        self.setLayout(QtWidgets.QHBoxLayout())
        label = QtWidgets.QLabel('!')
        label.setAlignment(QtCore.Qt.AlignCenter)
        label.setToolTip(tooltip)
        self.setToolTip(tooltip)
        self.layout().addWidget(label)
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.setMinimumSize(QtWidgets.qApp.fontMetrics().size(QtCore.Qt.TextSingleLine, 'OpenDXMC'))

        self.setVisible(False)

    @QtCore.pyqtSlot()
    def progress(self):
        self.progress = (self.progress + 64) % 5760

    @QtCore.pyqtSlot(bool)
    def busy(self, start):
        self.setMinimumSize(QtWidgets.qApp.fontMetrics().size(QtCore.Qt.TextSingleLine, '!!!')*2)
        if start and not self.isVisible():
            self.timer.start(50)
            self.show()
        elif not start:
            self.hide()
            self.timer.stop()

    def paintEvent(self, ev):
        if self.width() > self.height():
            d = self.height()
        else:
            d = self.width()

        rect = QtCore.QRectF(0, 0, d*.70, d*.70)
        rect.moveCenter(QtCore.QPointF(self.rect().center()))

        self.pen.setWidthF(d * .15)
        p = QtGui.QPainter(self)
        p.setRenderHint(p.Antialiasing, True)

        self.pen.setColor(QtGui.QColor.fromHsv((self.progress // 16) % 180 + 0,
                                               255, 255))
        p.setPen(self.pen)
        p.drawArc(rect, self.progress, 960)
        self.pen.setColor(QtGui.QColor.fromHsv((self.progress // 16) % 180 + 180,
                                               255, 255))
        p.setPen(self.pen)
        p.drawArc(rect, self.progress + 2880, 960)


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()

        self.setWindowIcon(QtGui.QIcon(os.path.join(os.path.dirname(__file__), 'icon.png')))
        self.setWindowIconText('OpenDXMC')
        self.setWindowTitle('OpenDXMC {}'.format(opendxmc.VERSION))        
        
        database_busywidget = BusyWidget(tooltip='Writing or Reading to Database')
        simulation_busywidget = BusyWidget(tooltip='Monte Carlo simulation in progress')
        importer_busywidget = BusyWidget(tooltip='Importing DICOM files')
        importer_phantoms_busywidget = BusyWidget(tooltip='Importing digital phantoms')

        mc_progressbar = RunnerView()

        # statusbar
        status_bar = QtWidgets.QStatusBar()
        statusbar_log_button = StatusBarButton('Log', None)
        status_bar.addPermanentWidget(importer_busywidget)
        status_bar.addPermanentWidget(importer_phantoms_busywidget)
        status_bar.addPermanentWidget(simulation_busywidget)
        status_bar.addPermanentWidget(database_busywidget)
        status_bar.addPermanentWidget(statusbar_log_button)

        self.setStatusBar(status_bar)

        # logging
        self.log_widget = LogWidget()
        self.log_handler = LogHandler(self)
        self.log_handler.message.connect(self.log_widget.insertPlainText)
        self.log_widget.closed.connect(statusbar_log_button.setChecked)
        statusbar_log_button.toggled.connect(self.log_widget.setVisible)
        logger.addHandler(self.log_handler)

        # central widget setup
        central_widget = QtWidgets.QWidget()
        central_widget.setContentsMargins(0, 0, 0, 0)
        self.setContentsMargins(0, 0, 0, 0)
        central_layout = QtWidgets.QHBoxLayout()
        central_splitter = QtWidgets.QSplitter(central_widget)
        central_layout.addWidget(central_splitter)
        central_layout.setContentsMargins(0, 0, 0, 0)


        # Databse interface
        database_selector_widget = SelectDatabaseWidget()
        self.interface = DatabaseInterface(database_selector_widget.path)
        self.interface.database_busy.connect(database_busywidget.busy)
        self.interface.database_busy.connect(database_selector_widget.locked)
        database_selector_widget.set_database_path.connect(self.interface.set_database)
        self.interface.send_proper_database_path.connect(database_selector_widget.validate_apply)



        # importer
        self.importer = Importer(self.interface)
        self.importer.running.connect(importer_busywidget.busy)
        self.importer.running.connect(database_selector_widget.locked)
        self.importer_phantom = Importer(self.interface)
        self.importer_phantom.running.connect(importer_phantoms_busywidget.busy)
        self.importer_phantom.running.connect(database_selector_widget.locked)

        ## import scaling setter
        import_scaling_widget = ImportScalingEdit(self.importer, self)

        self.viewcontroller = ViewController(self.interface)


        ## MC runner
        self.mcrunner = RunManager(self.interface, mc_progressbar)
        self.mcrunner.mc_calculation_running.connect(simulation_busywidget.busy)
        self.mcrunner.mc_calculation_running.connect(database_selector_widget.locked)


        # Models
        self.simulation_list_model = ListModel(self.interface, self.importer, self.importer_phantom, self,
                                               simulations=True)
        simulation_list_view = ListView()
        simulation_list_view.setModel(self.simulation_list_model)
        self.simulation_list_model.request_viewing.connect(self.viewcontroller.set_simulation)



        self.material_list_model = ListModel(self.interface, self,
                                             materials=True)
        self.material_list_model.request_viewing.connect(self.interface.emit_material_for_viewing)
        material_list_view = ListView()
        material_list_view.setModel(self.material_list_model)

        # Widgets

        list_view_collection_widget = QtWidgets.QWidget()
        list_view_collection_widget.setContentsMargins(0, 0, 0, 0)
        list_view_collection_widget.setLayout(QtWidgets.QVBoxLayout())
        list_view_collection_widget.layout().setContentsMargins(0, 0, 0, 0)
        list_view_collection_widget.layout().addWidget(import_scaling_widget, 1)
        list_view_collection_widget.layout().addWidget(database_selector_widget, 1)
        list_view_collection_widget.layout().addWidget(simulation_list_view, 300)
        list_view_collection_widget.layout().addWidget(material_list_view, 100)
        central_splitter.addWidget(list_view_collection_widget)

        properties_collection_widget = QtWidgets.QWidget()
        properties_collection_widget.setContentsMargins(0, 0, 0, 0)
        properties_collection_widget.setLayout(QtWidgets.QVBoxLayout())
        properties_collection_widget.layout().setContentsMargins(0, 0, 0, 0)
        properties_collection_widget.layout().addWidget(import_scaling_widget, 1)
        
        simulation_editor = PropertiesEditWidget(self.interface, self.simulation_list_model, self.mcrunner)
        self.viewcontroller.set_simulation_editor(simulation_editor.model)
        properties_collection_widget.layout().addWidget(simulation_editor, 3)
        properties_collection_widget.layout().addWidget(mc_progressbar, 1)
        central_splitter.addWidget(properties_collection_widget)

        for wwid in self.viewcontroller.view_widget():
            central_splitter.addWidget(wwid)

        self.organdosemodel = OrganDoseModel(self.interface, self.simulation_list_model)
        organdoseview = OrganDoseView(self.organdosemodel)
        central_splitter.addWidget(organdoseview)

        central_splitter.setSizes([100, 100, 100, 10, 0])

        central_widget.setLayout(central_layout)
        self.setCentralWidget(central_widget)

        # threading
        self.database_thread = QtCore.QThread(self)
        self.interface.moveToThread(self.database_thread)

#        self.mc_thread = QtCore.QThread(self)
#        self.mcrunner.moveToThread(self.mc_thread)

        self.dose_thread = QtCore.QThread(self)
        self.organdosemodel.moveToThread(self.dose_thread)

        self.import_thread = QtCore.QThread(self)
        self.importer.moveToThread(self.import_thread)
        self.import_phantom_thread = QtCore.QThread(self)
        self.importer_phantom.moveToThread(self.import_phantom_thread)
#        self.importer.moveToThread(self.database_thread)

        self.import_thread.start()
        self.import_phantom_thread.start()
#        self.mc_thread.start()
        self.dose_thread.start()
        self.database_thread.start()

        self.mcrunner.runner.finished.emit()





# -*- coding: utf-8 -*-
"""
Created on Tue Jun 23 13:47:11 2015

@author: ERLEAN
"""

import os
import math
import numpy as np

def rebin_scaling(a, scaling):
    factor_shape = tuple( s - (s % scaling[i])  for i, s in enumerate(a.shape))
    a_n = a[:factor_shape[0], :factor_shape[1], :factor_shape[2]]
    return rebin(a_n, scaling)


def rebin(a, factor):
    '''rebin ndarray data into a smaller ndarray of the same rank whose dimensions
    are factors of the original dimensions. eg. An array with 6 columns and 4 rows
    can be reduced to have 6,3,2 or 1 columns and 4,2 or 1 rows.
    example usages:
    >>> a=rand(6,4); b=rebin(a,3,2)
    >>> a=rand(6); b=rebin(a,2)
    '''
    shape = a.shape
    lenShape = len(shape)
    factor = np.asarray(factor, dtype=np.int)
    args = (np.asarray(shape) / factor).astype(np.int)
    evList = ['a.reshape('] + \
             ['args[{}],factor[{}],'.format(i,i) for i in range(lenShape)] + \
             [')'] + ['.sum({})'.format(i+1) for i in range(lenShape)] + \
             ['//factor[{}]'.format(i) for i in range(lenShape)]
    return eval(''.join(evList))


def human_time(sec):
    if sec < 30:
        return "a moment"
    elif sec < 120:
        return 'a minute or so'
    elif sec < 3600:
        return "about {} minutes".format(int(sec) // 60)
    elif sec < 3600*2:
        minutes = (int(sec) - 3600) // 60
        if minutes < 5:
            return "about an hour"
        return "one hour and {} minutes".format(minutes)
    elif sec < 3600*3:
        minutes = (int(sec) - 3600*2) // 60
        if minutes < 5:
            return "about two hours"
        return "two hours and {} minutes".format(minutes)
    
    hours = int(sec) // 3600
    return "about {} hours".format(hours)        

#def human_time(sec):
#    seconds = [86400., 3600., 60., 1.]
#    names = ['days', 'hours', 'minutes', 'seconds']
#
#    if sec <= seconds[-1]:
#        return 'less than a {0}'.format(names[-1][:-1])
#    times = []
#    labels = []
#    for i, s in enumerate(seconds):
#        if sec >= s:
#            times.append(int(math.floor(sec/s)))
#            labels.append(names[i])
#            if times[0] == 1:
#                labels[0] = labels[0][:-1]
#            if i < len(seconds) - 1:
#                times.append(int(math.floor((sec - s * times[0])/seconds[i+1])))
#                labels.append(names[i+1])
#                if times[1] == 1:
#                    labels[1] = labels[1][:-1]
#
#            break
#    else:
#        return 'less than a {0}'.format(names[-1][:-1])
#    if len(times) > 1:
#        if times[-1] == 0:
#            return 'about {0} {1}'.format(times[0], labels[0])
#        return 'about {0} {1} and {2} {3}'.format(times[0], labels[0], times[1], labels[1])
#    return 'about {0} {1}'.format(times[0], labels[0])


def circle_mask(array_shape, radius, center=None):
    a = np.zeros(array_shape, np.int)
    if not center:
        cx = array_shape[0] // 2
        cy = array_shape[1] // 2
    else:
        cx, cy = (int(c) for c in center)
    rint = int(np.ceil(radius))
    y, x = np.ogrid[-rint: rint+1, -rint: rint+1]
    index = x**2 + y**2 <= radius**2
    a[cx-rint:cx+rint+1, cy-rint:cy+rint+1][index] = 1
    return a


def sphere_mask(array_shape, radius, center=None):
    a = np.zeros(array_shape, np.int)
    if not center:
        cx = array_shape[0] / 2
        cy = array_shape[1] / 2
        cz = array_shape[2] / 2
    else:
        cx, cy, cz = center
    x, y, z = np.ogrid[-radius: radius, -radius: radius, -radius: radius]
    index = x**2 + y**2 + z**2 <= radius**2
    a[cx-radius:cx+radius+1, cy-radius:cy+radius+1, cz-radius:cz+radius+1][index] = 1
    return a


def find_all_files(pathList):
    for p in pathList:
        path = os.path.abspath(p)
        if os.path.isdir(path):
            for dirname, dirnames, filenames in os.walk(path):
                for filename in filenames:
                    yield os.path.normpath(os.path.join(dirname, filename))
        elif os.path.isfile(path):
            yield os.path.normpath(path)


